﻿/******************************************************************************
Send bug reports to: philip@spigelmire.com

Copyright (c) 2010, Weatherite Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation 
and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
******************************************************************************/

using System;
using System.Text;
using System.Net.Sockets;

namespace Opto22Library
{
    public class O22SnapIoMemMap : IDisposable
    {
        //public delegate void ConnectFinishedEventHandler(object source, O22SnapIoEventArgs ea);
        //public event ConnectFinishedEventHandler ConnectFinished;

        public O22SnapIoMemMap()
        {

            // Set defaults
            m_byTransactionLabel = 1;
        }

        private byte m_byTransactionLabel;
        private bool m_nAutoPUCFlag;
        private ConnectionType m_nConnectionType;
        private Socket m_Socket;

        public void Dispose()
        {
            CloseSockets();
        }

        //Caller should register and wait for ConnectFinished event.
        public void Open(string pchIpAddressArg, int nPort, int nOpenTimeOutMS, bool nAutoPUC)
        {
            Open(pchIpAddressArg, nPort, nOpenTimeOutMS, nAutoPUC, ConnectionType.TCP);
        }

        public void Open(string Host, int Port, int OpenTimeOutMS, bool AutoPUC, ConnectionType ConnectionType)
        {
            m_nAutoPUCFlag = AutoPUC;
            m_nConnectionType = ConnectionType;

            OpenSockets(Host, Port, OpenTimeOutMS);
        }

        private void OpenSockets(string Host, int Port, int OpenTimeOutMS)
        //-------------------------------------------------------------------------------------------------
        // Use sockets to open a connection to the SNAP I/O unit
        //-------------------------------------------------------------------------------------------------
        {
            // If a socket is open, close it now.
            CloseSockets();

            SocketType socketType;
            if (m_nConnectionType == ConnectionType.TCP)
                socketType = SocketType.Stream;
            else
                socketType = SocketType.Dgram;

            // Create the socket via TCP or UDP
            m_Socket = new Socket(AddressFamily.InterNetwork, socketType, ProtocolType.IP);
            // Make the socket non-blocking - we can call the asynch methods for this
            m_Socket.Blocking = true;  //we will do the nonblocking through async methods

            m_Socket.SendTimeout = OpenTimeOutMS;
            m_Socket.ReceiveTimeout = OpenTimeOutMS;
            // attempt connection
            //m_Socket.Connect(pchIpAddressArg, nPort);  //This will throw an exception on TCP since blocking is false.
            //m_Socket.BeginConnect(Host, Port, new AsyncCallback(ConnectCallback), m_Socket);  //What is the advantage of doing this vs. blocking and catching timeout exceptions?

            //Without the callback
            m_Socket.Connect(Host, Port);

            if (m_nAutoPUCFlag)
                if (GetStatusPUC())
                    SetStatusOperation(SIOMM_OpCode.PowerUpClear);
        }

        private void UnpackWriteResponse(byte[] byWriteResponse, out byte byTransactionLabel, out byte byResponseCode)
        //-------------------------------------------------------------------------------------------------
        // Unpack a write response from the brain
        //-------------------------------------------------------------------------------------------------
        {
            byTransactionLabel = 0;
            byResponseCode = 0;

            // Check for correct Transaction Code
            byte byTransactionCode = (byte)(byWriteResponse[3] >> 4);
            if ((byte)TransactionCode.SIOMM_TCODE_WRITE_RESPONSE != byTransactionCode)
                throw new Opto22Library.O22SnapIoException(SIOMM.ERROR_BAD_TCODE);

            // Unpack Transaction Label
            byTransactionLabel = byWriteResponse[2];
            byTransactionLabel >>= 2;

            // Unpack Response Code
            byResponseCode = byWriteResponse[6];
            byResponseCode >>= 4;
        }

        private byte [] BuildWriteQuadletRequest(byte TransactionLabel, Int16 SourceId, UInt32 DestinationOffset, UInt32 QuadletData)
        //-------------------------------------------------------------------------------------------------
        // Build a write quadlet request packet
        //-------------------------------------------------------------------------------------------------
        {
            
            byte[] WriteQuadletRequest = new byte[SIOMM.SIZE_WRITE_QUAD_REQUEST];

            WriteQuadletRequest[0] = 0x00;
            WriteQuadletRequest[1] = 0x00;
            WriteQuadletRequest[2] = (byte)(TransactionLabel << 2);
            WriteQuadletRequest[3] = (byte)TransactionCode.SIOMM_TCODE_WRITE_QUAD_REQUEST << 4;
            WriteQuadletRequest[4] = O22byte((UInt32)SourceId, 1); //Hi byte
            WriteQuadletRequest[5] = O22byte((UInt32)SourceId, 0); //Lo byte
            WriteQuadletRequest[6] = 0xFF;
            WriteQuadletRequest[7] = 0xFF;
            WriteQuadletRequest[8] = O22byte(DestinationOffset, 3);
            WriteQuadletRequest[9] = O22byte(DestinationOffset, 2);
            WriteQuadletRequest[10] = O22byte(DestinationOffset, 1);
            WriteQuadletRequest[11] = O22byte(DestinationOffset, 0);
            WriteQuadletRequest[12] = O22byte(QuadletData, 3);
            WriteQuadletRequest[13] = O22byte(QuadletData, 2);
            WriteQuadletRequest[14] = O22byte(QuadletData, 1);
            WriteQuadletRequest[15] = O22byte(QuadletData, 0);

            return WriteQuadletRequest;

        }

        private byte[] BuildWriteBlockRequest(byte TransactionLabel, UInt32 DestinationOffset, byte[] BlockData)
        //-------------------------------------------------------------------------------------------------
        // Build a write block request packet
        //-------------------------------------------------------------------------------------------------
        {
            UInt16 DataLength = (UInt16)BlockData.Length;

            if (DataLength > 2034 && m_Socket.SocketType == SocketType.Stream
                || DataLength > 1480 && m_Socket.SocketType == SocketType.Dgram)
                throw new Opto22Library.O22SnapIoException(SIOMM.ERROR_PAYLOAD_TOO_BIG);

            byte[] byWriteBlockRequest = new byte[SIOMM.SIZE_WRITE_BLOCK_REQUEST + DataLength];

            // Destination Id
            byWriteBlockRequest[0] = 0x00;
            byWriteBlockRequest[1] = 0x00;

            // Transaction Label
            byWriteBlockRequest[2] = (byte)(TransactionLabel << 2);

            // Transaction Code
            byWriteBlockRequest[3] = (byte)TransactionCode.SIOMM_TCODE_WRITE_BLOCK_REQUEST << 4;

            // Source Id
            byWriteBlockRequest[4] = 0x00;
            byWriteBlockRequest[5] = 0x00;

            // Destination Offset
            byWriteBlockRequest[6] = 0xFF;
            byWriteBlockRequest[7] = 0xFF;
            byWriteBlockRequest[8] = O22byte(DestinationOffset, 3);
            byWriteBlockRequest[9] = O22byte(DestinationOffset, 2);
            byWriteBlockRequest[10] = O22byte(DestinationOffset, 1);
            byWriteBlockRequest[11] = O22byte(DestinationOffset, 0);

            // Data length
            byWriteBlockRequest[12] = O22byte(DataLength, 1);
            byWriteBlockRequest[13] = O22byte(DataLength, 0);

            // Extended Transaction Code
            byWriteBlockRequest[14] = 0x00;
            byWriteBlockRequest[15] = 0x00;

            // Block data to write
            System.Array.Copy(BlockData, 0, byWriteBlockRequest, 16, DataLength);
            //memcpy(&(byWriteBlockRequest[16]), byBlockData, wDataLength);

            return byWriteBlockRequest;
        }

        private void CloseSockets()
        //-------------------------------------------------------------------------------------------------
        // Close the sockets connection
        //-------------------------------------------------------------------------------------------------
        {
            // Close up everything
            if (m_Socket != null)
                m_Socket.Close();

            // Reset our data members
            m_Socket = null;
        }

        private UInt32 ReadQuad(UInt32 DestOffset)
        {
            byte[] byReadQuadletRequest;
            byte[] byReadQuadletResponse = new byte[SIOMM.SIZE_READ_QUAD_RESPONSE];
            byte byTransactionLabel;
            byte byResponseCode;
            UInt32 QuadletTemp = 0;

            if (m_Socket == null)
                throw new Opto22Library.O22SnapIoException(SIOMM.ERROR_NOT_CONNECTED);

            // Increment the transaction label
            UpdateTransactionLabel();

            // Build the request packet
            byReadQuadletRequest =BuildReadQuadletRequest(m_byTransactionLabel, DestOffset);

            // Send the packet to the Snap I/O unit
            m_Socket.Send(byReadQuadletRequest);

            m_Socket.Receive(byReadQuadletResponse);

            // Unpack the response.
            QuadletTemp = UnpackReadQuadletResponse(byReadQuadletResponse,
                                                out byTransactionLabel,
                                                out byResponseCode);


            if(m_byTransactionLabel != byTransactionLabel)
                throw new Opto22Library.O22SnapIoException(SIOMM.ERROR_RESPONSE_BAD);

            // Check that the response was okay
            if (SIOMM.RESPONSE_CODE_ACK != byResponseCode)
            {
                // If a bad response from the brain, get its last error code
                int nErrorCode = GetStatusLastError();
                throw new Opto22Library.O22SnapIoException(nErrorCode);
            }

            return QuadletTemp;

        }

        private void WriteQuad(UInt32 DestOffset, UInt32 Quadlet)
        //-------------------------------------------------------------------------------------------------
        // Write a quadlet of data to a location in the SNAP I/O memory map.
        //-------------------------------------------------------------------------------------------------
        {
            byte[] byWriteQuadletRequest;
            byte[] byWriteQuadletResponse = new byte[SIOMM.SIZE_WRITE_RESPONSE];
            byte byTransactionLabel = 0;
            byte byResponseCode = 0;
            int nResult;
            //fd_set fds;

            // Check that we have a valid socket
            if (m_Socket == null)
                throw new Opto22Library.O22SnapIoException(SIOMM.ERROR_NOT_CONNECTED);

            // Increment the transaction label
            UpdateTransactionLabel();

            // Build the write request packet
            byWriteQuadletRequest = BuildWriteQuadletRequest(m_byTransactionLabel, 1,
                                     DestOffset, Quadlet);


            // Send the packet to the Snap I/O unit
            m_Socket.Send(byWriteQuadletRequest);

            nResult = m_Socket.Receive(byWriteQuadletResponse);

            // Check if we got the wrong number of bytes back!
            if (SIOMM.SIZE_WRITE_RESPONSE != nResult)
                throw new Opto22Library.O22SnapIoException(SIOMM.ERROR_RESPONSE_BAD);

            // Unpack the response
            UnpackWriteResponse(byWriteQuadletResponse, out byTransactionLabel, out byResponseCode);

            //Make sure Transaction Label matches
            if (m_byTransactionLabel != byTransactionLabel)
                throw new Opto22Library.O22SnapIoException(SIOMM.ERROR_RESPONSE_BAD);

            // Check that the device response was okay
            if (SIOMM.RESPONSE_CODE_ACK != byResponseCode)
            {
                // If a bad response from the brain, get its last error code
                int ErrorCode = GetStatusLastError();
                throw new Opto22Library.O22SnapIoException(ErrorCode);
            }
        }

        private byte[] ReadBlock(UInt32 DestOffset, UInt32 PayloadLength)
        //-------------------------------------------------------------------------------------------------
        // Read a block of data from a location in the SNAP I/O memory map.
        //-------------------------------------------------------------------------------------------------
        {
            byte[] byReadBlockRequest;
            byte[] byReadBlockResponse = new byte[PayloadLength + SIOMM.SIZE_READ_BLOCK_RESPONSE];
            byte[] Payload;
            byte byTransactionLabel;
            byte byResponseCode;
            UInt32 MaxPayloadSize;
            byte[] PayloadChunk;
            UInt32 PayloadCount;
            UInt16 ChunkLength;

            // Check that we have a valid socket
            if (m_Socket == null)
                throw new Opto22Library.O22SnapIoException(SIOMM.ERROR_NOT_CONNECTED);

            Payload = new byte[PayloadLength];

            //Loop the read requests for large payload requests (at 1480 byte boundaries for UDP, 2034 for TCP)
            if (m_Socket.SocketType == SocketType.Stream)
                //Ethernet frame 1500 bytes, 54 bytes overhead, leaves 1446 for data, 1444 on 4 byte boundery, less 16 bytes of OptoMMP overhead = 1428 - may want less if packets are getting encapsulated like in a VPN
                //MaxPayloadSize = 2032;  //4 byte boundary1444
                MaxPayloadSize = 1436;
            else
                MaxPayloadSize = 1456;

            PayloadCount = PayloadLength / MaxPayloadSize;
            if(PayloadLength % MaxPayloadSize > 0)
                PayloadCount++;

            for (UInt16 i = 0; i < PayloadCount; i++)
            {
                if (i < PayloadCount - 1) //request full payload
                    ChunkLength = (UInt16)MaxPayloadSize;
                else
                    ChunkLength = (UInt16)(PayloadLength - (i * MaxPayloadSize));

                // Increment the transaction label
                UpdateTransactionLabel();

                // Build the request packet
                byReadBlockRequest = BuildReadBlockRequest(m_byTransactionLabel, DestOffset + (uint)(i * MaxPayloadSize), ChunkLength);

                // Send the packet to the Snap I/O unit
                //nResult = send(m_Socket, (char*)byReadBlockRequest, SIOMM_SIZE_READ_BLOCK_REQUEST, 0 /*??*/ );
                m_Socket.Send(byReadBlockRequest);
                int ResultLength = m_Socket.Receive(byReadBlockResponse);

                if ((ChunkLength + SIOMM.SIZE_READ_BLOCK_RESPONSE) != ResultLength)
                    throw new Opto22Library.O22SnapIoException(SIOMM.ERROR_RESPONSE_BAD);

                // Unpack the response.
                PayloadChunk = UnpackReadBlockResponse(byReadBlockResponse, out byTransactionLabel, out byResponseCode, ChunkLength);

                //Make sure Transaction Label matches
                if (m_byTransactionLabel != byTransactionLabel)
                    throw new Opto22Library.O22SnapIoException(SIOMM.ERROR_RESPONSE_BAD);

                // Check that the device response was okay
                if (SIOMM.RESPONSE_CODE_ACK != byResponseCode)
                {
                    // If a bad response from the brain, get its last error code
                    int ErrorCode = GetStatusLastError();
                    throw new Opto22Library.O22SnapIoException(ErrorCode);
                }

                PayloadChunk.CopyTo(Payload, i * MaxPayloadSize);
            }


            return Payload;
        }

        private void WriteBlock(UInt32 DestOffset, byte[] Payload)
        //-------------------------------------------------------------------------------------------------
        // Write a block of data to a location in the SNAP I/O memory map.
        //-------------------------------------------------------------------------------------------------
        {
            byte[] byWriteBlockRequest;
            byte[] byWriteBlockResponse = new byte[SIOMM.SIZE_WRITE_RESPONSE];
            UInt16 PayloadLength = (UInt16)Payload.Length;
            byte byTransactionLabel;
            byte byResponseCode;
            int MaxPayloadSize;
            byte [] PayloadChunk;

            // Check that we have a valid socket
            if (m_Socket == null)
                throw new Opto22Library.O22SnapIoException(SIOMM.ERROR_NOT_CONNECTED);

            //Loop the writing for large payloads (at 1480 byte boundaries for UDP, 2034 for TCP)
            if (m_Socket.SocketType == SocketType.Stream)
                MaxPayloadSize = 2032; //2034 is max, 2032 for 4 byte boundary
            else
                MaxPayloadSize = 1480;

            for (int i = 0; i < Payload.Length; i += MaxPayloadSize)
            {
                // Increment the transaction label
                UpdateTransactionLabel();

                if (Payload.Length - i < MaxPayloadSize)
                    PayloadChunk = new byte[Payload.Length - i];
                else
                    PayloadChunk = new byte[MaxPayloadSize];
                
                Array.Copy(Payload, i, PayloadChunk, 0, PayloadChunk.Length);

                // Build the write request packet
                byWriteBlockRequest = BuildWriteBlockRequest(m_byTransactionLabel, (UInt32)(DestOffset + i), PayloadChunk);

                // Send the packet to the Snap I/O unit
                //nResult = send(m_Socket, (char*)pbyWriteBlockRequest, SIOMM_SIZE_WRITE_BLOCK_REQUEST + wDataLength, 0 /*??*/ );
                m_Socket.Send(byWriteBlockRequest);

                // The response is ready, so recv it.
                //nResult = recv(m_Socket, (char*)byWriteBlockResponse, SIOMM_SIZE_WRITE_RESPONSE, 0 /*??*/);
                int ResultLength = m_Socket.Receive(byWriteBlockResponse);

                if (SIOMM.SIZE_WRITE_RESPONSE != ResultLength)
                    throw new Opto22Library.O22SnapIoException(SIOMM.ERROR_RESPONSE_BAD);

                // Unpack the response
                UnpackWriteResponse(byWriteBlockResponse, out byTransactionLabel,
                                              out byResponseCode);

                //Make sure Transaction Label matches
                if (m_byTransactionLabel != byTransactionLabel)
                    throw new Opto22Library.O22SnapIoException(SIOMM.ERROR_RESPONSE_BAD);

                // Check that the device response was okay
                if (SIOMM.RESPONSE_CODE_ACK != byResponseCode)
                {
                    // If a bad response from the brain, get its last error code
                    int ErrorCode = GetStatusLastError();
                    throw new Opto22Library.O22SnapIoException(ErrorCode);
                }
            }


        }

        private UInt32 UnpackReadQuadletResponse(byte[] ReadQuadletResponse,
                out byte TransactionLabel, out byte ResponseCode)
        {
            //-------------------------------------------------------------------------------------------------
            // Unpack a read quadlet response from the brain
            //-------------------------------------------------------------------------------------------------
            UInt32 QuadletData;

            // Check for correct Transaction Code
            byte TransCode = (byte)(ReadQuadletResponse[3] >> 4);
            if ((byte)TransactionCode.SIOMM_TCODE_READ_QUAD_RESPONSE != TransCode)
                throw new Opto22Library.O22SnapIoException(SIOMM.ERROR_BAD_TCODE);

            TransactionLabel = ReadQuadletResponse[2];
            TransactionLabel >>= 2;

            ResponseCode = ReadQuadletResponse[6];
            ResponseCode >>= 4;

            //QuadletData = O22MAKELONG(byReadQuadletResponse[12], byReadQuadletResponse[13], byReadQuadletResponse[14], byReadQuadletResponse[15]);
            QuadletData = O22SnapIoHelper.ToUInt32(ReadQuadletResponse, 12);

            return QuadletData;
        }

        private byte O22byte(UInt32 word, int pos)
        {
            //Return the byte from a uint32 at position pos.
            //pos is 0,1,2, or 3
            // O22byte0(l) ((byte)(((DWORD)(l) >> 24) & 0xFF))  opposite of this so it works with different word sizes
            return (byte)(word >> (pos * 8) & 0xFF);
        }

        private byte [] BuildReadQuadletRequest(byte TransactionLabel, UInt32 DestOffset)
        //-------------------------------------------------------------------------------------------------
        // Build a read quadlet request packet
        //-------------------------------------------------------------------------------------------------
        {
            byte[] ReadQuadletRequest = new byte[SIOMM.SIZE_READ_QUAD_REQUEST];

            ReadQuadletRequest[0] = 0x00;
            ReadQuadletRequest[1] = 0x00;
            ReadQuadletRequest[2] = (byte)(TransactionLabel << 2);
            ReadQuadletRequest[3] = (byte)TransactionCode.SIOMM_TCODE_READ_QUAD_REQUEST << 4;
            ReadQuadletRequest[4] = 0x00;
            ReadQuadletRequest[5] = 0x00;
            ReadQuadletRequest[6] = 0xFF;
            ReadQuadletRequest[7] = 0xFF;
            ReadQuadletRequest[8] = O22byte(DestOffset, 3);
            ReadQuadletRequest[9] = O22byte(DestOffset, 2);
            ReadQuadletRequest[10] = O22byte(DestOffset, 1);
            ReadQuadletRequest[11] = O22byte(DestOffset, 0);

            return ReadQuadletRequest;
        }

        private byte[] BuildReadBlockRequest(byte TransactionLabel, UInt32 DestinationOffset, UInt16 DataLength)
        //-------------------------------------------------------------------------------------------------
        // Build a read block request packet
        //-------------------------------------------------------------------------------------------------
        {
            byte[] ReadBlockRequest = new byte[SIOMM.SIZE_READ_BLOCK_REQUEST];

            // Destination Id
            ReadBlockRequest[0] = 0x00;
            ReadBlockRequest[1] = 0x00;

            // Transaction Label
            ReadBlockRequest[2] = (byte)(TransactionLabel << 2);

            // Transaction Code
            ReadBlockRequest[3] = (byte)TransactionCode.SIOMM_TCODE_READ_BLOCK_REQUEST << 4;

            // Source Id
            ReadBlockRequest[4] = 0x00;
            ReadBlockRequest[5] = 0x00;

            // Destination Offset
            ReadBlockRequest[6] = 0xFF;
            ReadBlockRequest[7] = 0xFF;
            ReadBlockRequest[8] = O22byte(DestinationOffset, 3);
            ReadBlockRequest[9] = O22byte(DestinationOffset, 2);
            ReadBlockRequest[10] = O22byte(DestinationOffset, 1);
            ReadBlockRequest[11] = O22byte(DestinationOffset, 0);

            // Data length
            ReadBlockRequest[12] = O22byte(DataLength, 1);
            ReadBlockRequest[13] = O22byte(DataLength, 0);

            // Extended Transaction Code
            ReadBlockRequest[14] = 0x00;
            ReadBlockRequest[15] = 0x00;

            return ReadBlockRequest;
        }

        // Gets the next transaction label
        private void UpdateTransactionLabel()
        {
            m_byTransactionLabel++;
            if (m_byTransactionLabel >= 64)
                m_byTransactionLabel = 0;
        }

        public void SetCommOptions(int TimeOutMS)
        //-------------------------------------------------------------------------------------------------
        // Set communication options
        //-------------------------------------------------------------------------------------------------
        {
            if (m_Socket != null)
            {
                m_Socket.SendTimeout = TimeOutMS;
                m_Socket.ReceiveTimeout = TimeOutMS;
            }
            else
                throw new Opto22Library.O22SnapIoException(SIOMM.ERROR_NOT_CONNECTED);
        }


        private byte[] UnpackReadBlockResponse(byte[] byReadBlockResponse, out byte byTransactionLabel, out byte byResponseCode, UInt16 PayloadLength)
        //-------------------------------------------------------------------------------------------------
        // Unpack a read block response from the brain
        //-------------------------------------------------------------------------------------------------
        {
            byte[] Payload = new byte[PayloadLength];
            UInt16 ResponsePayloadLength;

            // Check for correct Transaction Code
            byte TransCode = (byte)(byReadBlockResponse[3] >> 4);
            if ((byte)TransactionCode.SIOMM_TCODE_READ_BLOCK_RESPONSE != TransCode)
                throw new Opto22Library.O22SnapIoException(SIOMM.ERROR_BAD_TCODE);

            byTransactionLabel = byReadBlockResponse[2];
            byTransactionLabel >>= 2;

            byResponseCode = byReadBlockResponse[6];
            byResponseCode >>= 4;

            ResponsePayloadLength = O22SnapIoHelper.ToUInt16(byReadBlockResponse, 12);

            if (ResponsePayloadLength != PayloadLength)
                throw new Opto22Library.O22SnapIoException(SIOMM.ERROR_RESPONSE_BAD);

            System.Array.Copy(byReadBlockResponse, 16, Payload, 0, PayloadLength);

            return Payload;
        }

        private Single ReadFloat(UInt32 DestOffset)
        //-------------------------------------------------------------------------------------------------
        // Read a float value from a location in the SNAP I/O memory map.
        //-------------------------------------------------------------------------------------------------
        {
            UInt32 Quadlet = ReadQuad(DestOffset);
            return O22SnapIoHelper.ToSingle(O22SnapIoHelper.GetBytes(Quadlet), 0);
        }

        private void WriteFloat(UInt32 DestOffset, Single Value)
        //-------------------------------------------------------------------------------------------------
        // Write a float value to a location in the SNAP I/O memory map.
        //-------------------------------------------------------------------------------------------------
        {
            // Copy the float into a DWORD for easy manipulation
            UInt32 Quadlet = O22SnapIoHelper.ToUInt32(O22SnapIoHelper.GetBytes(Value), 0);
            WriteQuad(DestOffset, Quadlet);
        }

        public bool GetDigPtState(short Point)
        //-------------------------------------------------------------------------------------------------
        // Get the state of the specified digital point.
        //-------------------------------------------------------------------------------------------------
        {
            UInt32 quad = ReadQuad(SIOMM.MM.POINT.DPOINT.READ_STATE + (UInt32)(SIOMM.MM.POINT.DPOINT.READ_BOUNDARY * Point));
            return (quad != 0);
        }

        public bool GetDigPtOnLatch(short Point)
        //-------------------------------------------------------------------------------------------------
        // Get the on-latch state of the specified digital point.
        //-------------------------------------------------------------------------------------------------
        {
            UInt32 quad = ReadQuad(SIOMM.MM.POINT.DPOINT.READ_ONLATCH_STATE + (UInt32)(SIOMM.MM.POINT.DPOINT.READ_BOUNDARY * Point));
            return (quad != 0);
        }

        public bool GetDigPtOffLatch(short Point)
        //-------------------------------------------------------------------------------------------------
        // Get the off-latch state of the specified digital point.
        //-------------------------------------------------------------------------------------------------
        {
            UInt32 quad = ReadQuad(SIOMM.MM.POINT.DPOINT.READ_OFFLATCH_STATE + (UInt32)(SIOMM.MM.POINT.DPOINT.READ_BOUNDARY * Point));
            return (quad != 0);
        }

        public bool GetDigPtCounterState(short Point)
        //-------------------------------------------------------------------------------------------------
        // Get the active counter state of the specified digital point.
        //-------------------------------------------------------------------------------------------------
        {
            UInt32 quad = ReadQuad(SIOMM.MM.POINT.DPOINT.READ_ACTIVE_COUNTER + (UInt32)(SIOMM.MM.POINT.DPOINT.READ_BOUNDARY * Point));
            return (quad != 0);
        }

        public UInt32 GetDigPtCounts(short Point)
        //-------------------------------------------------------------------------------------------------
        // Get the counters of the specified digital point.
        //-------------------------------------------------------------------------------------------------
        {
            return ReadQuad(SIOMM.MM.POINT.DPOINT.READ_COUNTER_DATA + (UInt32)(SIOMM.MM.POINT.DPOINT.READ_BOUNDARY * Point));
        }

        public SIOMM_DigPointReadArea GetDigPtReadAreaEx(short Point)
        //-------------------------------------------------------------------------------------------------
        // Get the read area for the specified digital point
        //-------------------------------------------------------------------------------------------------
        {
            SIOMM_DigPointReadArea Data = new SIOMM_DigPointReadArea();

            // Read the data
            byte[] arrbyData = ReadBlock(SIOMM.MM.POINT.DPOINT.READ_AREA_BASE + (UInt32)(SIOMM.MM.POINT.DPOINT.READ_BOUNDARY * Point), 20);

            // If everything is okay, go ahead and fill the structure
            Data.State = O22SnapIoHelper.ToBoolean(arrbyData, 0 + 3);
            Data.OnLatch = O22SnapIoHelper.ToBoolean(arrbyData, 4 + 3);
            Data.OffLatch = O22SnapIoHelper.ToBoolean(arrbyData, 8 + 3);
            Data.CounterState = O22SnapIoHelper.ToBoolean(arrbyData, 12 + 3);
            Data.Counts = O22SnapIoHelper.ToUInt32(arrbyData, 16);

            return Data;
            //pData->nState        = O22MAKELONG2(arrbyData, 0);
            //pData->nOnLatch      = O22MAKELONG2(arrbyData, 4);
            //pData->nOffLatch     = O22MAKELONG2(arrbyData, 8);
            //pData->nCounterState = O22MAKELONG2(arrbyData, 12);
            //pData->nCounts       = O22MAKELONG2(arrbyData, 16);
        }


        public void SetDigPtState(short Point, bool State)
        //-------------------------------------------------------------------------------------------------
        // Set the state of the specified digital point
        //-------------------------------------------------------------------------------------------------
        {
            if (State)
                WriteQuad(SIOMM.MM.POINT.DPOINT.WRITE_TURN_ON_BASE + (UInt32)(SIOMM.MM.POINT.DPOINT.WRITE_BOUNDARY * Point), 1);
            else
                WriteQuad(SIOMM.MM.POINT.DPOINT.WRITE_TURN_OFF_BASE + (UInt32)(SIOMM.MM.POINT.DPOINT.WRITE_BOUNDARY * Point), 1);
        }

        public void SetDigPtCounterState(short Point, bool Activate)
        //-------------------------------------------------------------------------------------------------
        // Set the active counter state of the specified digital point
        //-------------------------------------------------------------------------------------------------
        {
            if (Activate)
                WriteQuad(SIOMM.MM.POINT.DPOINT.WRITE_ACTIVATE_COUNTER + (UInt32)(SIOMM.MM.POINT.DPOINT.WRITE_BOUNDARY * Point), 1);
            else
                WriteQuad(SIOMM.MM.POINT.DPOINT.WRITE_DEACTIVATE_COUNTER + (UInt32)(SIOMM.MM.POINT.DPOINT.WRITE_BOUNDARY * Point), 1);
        }

        private UInt64 GetBitmask64(UInt32 DestOffset)
        //-------------------------------------------------------------------------------------------------
        // Get a 64-bit bitmask from a location in the SNAP I/O memory map.
        //-------------------------------------------------------------------------------------------------
        {
            // Read the data
            byte [] arrbyData = ReadBlock(DestOffset, 8);
            return O22SnapIoHelper.ToUInt64(arrbyData, 0);
        }

        private void SetBitmask64(UInt32 DestOffset, UInt64 Points)
        //-------------------------------------------------------------------------------------------------
        // Set a 64-bit bitmask to a location in the SNAP I/O memory map.
        //-------------------------------------------------------------------------------------------------
        {
            // Write the data
            WriteBlock(DestOffset, O22SnapIoHelper.GetBytes(Points));
        }

        public UInt64 GetDigBankPointStates()
        //-------------------------------------------------------------------------------------------------
        // Get the digital point states in the digital bank
        //-------------------------------------------------------------------------------------------------
        {
            return GetBitmask64(SIOMM.MM.DBANK.READ_POINT_STATES);
        }

        public UInt64 GetDigBankOnLatchStates()
        //-------------------------------------------------------------------------------------------------
        // Get the digital point on-latch states in the digital bank
        //-------------------------------------------------------------------------------------------------
        {
            return GetBitmask64(SIOMM.MM.DBANK.READ_ON_LATCH_STATES);
        }


        public UInt64 GetDigBankOffLatchStates()
        //-------------------------------------------------------------------------------------------------
        // Get the digital point off-latch states in the digital bank
        //-------------------------------------------------------------------------------------------------
        {
            return GetBitmask64(SIOMM.MM.DBANK.READ_OFF_LATCH_STATES);
        }


        public UInt64 GetDigBankActCounterStates()
        //-------------------------------------------------------------------------------------------------
        // Get the digital point active counter states in the digital bank
        //-------------------------------------------------------------------------------------------------
        {
            return GetBitmask64(SIOMM.MM.DBANK.READ_ACTIVE_COUNTERS);
        }


        public void SetDigBankPointStates(UInt64 Points, UInt64 Mask)
        //-------------------------------------------------------------------------------------------------
        // Set the digital point states in the digital bank.  Only those points set in the mask parameters
        // are set.
        //-------------------------------------------------------------------------------------------------
        {
            UInt64 OnMask, OffMask;
            byte[] Data = new byte[16];

            OnMask = Points & Mask;
            OffMask = (Points ^ UInt64.MaxValue) & Mask;

            Array.Copy(O22SnapIoHelper.GetBytes(OnMask), Data, 8);
            Array.Copy(O22SnapIoHelper.GetBytes(OffMask), 0, Data, 8, 8);

            WriteBlock(SIOMM.MM.DBANK.WRITE_TURN_ON_MASK, Data);
        }

        public void SetDigBankOnMask(UInt64 Points)
        //-------------------------------------------------------------------------------------------------
        // Turn on the given digital points
        //-------------------------------------------------------------------------------------------------
        {
            SetBitmask64(SIOMM.MM.DBANK.WRITE_TURN_ON_MASK, Points);
        }

        public void SetDigBankOffMask(UInt64 Points)
        //-------------------------------------------------------------------------------------------------
        // Turn off the given digital points
        //-------------------------------------------------------------------------------------------------
        {
            SetBitmask64(SIOMM.MM.DBANK.WRITE_TURN_OFF_MASK, Points);
        }

        public void SetDigBankActCounterMask(UInt64 Points)
        //-------------------------------------------------------------------------------------------------
        // Active counters for the given digital points
        //-------------------------------------------------------------------------------------------------
        {
            SetBitmask64(SIOMM.MM.DBANK.WRITE_ACT_COUNTERS_MASK, Points);
        }

        public void SetDigBankDeactCounterMask(UInt64 Points)
        //-------------------------------------------------------------------------------------------------
        // Deactive counters for the given digital points
        //-------------------------------------------------------------------------------------------------
        {
            SetBitmask64(SIOMM.MM.DBANK.WRITE_DEACT_COUNTERS_MASK, Points);
        }


        public SIOMM_DigBankReadArea GetDigBankReadAreaEx()
        //-------------------------------------------------------------------------------------------------
        // Get the read area for the digital bank
        //-------------------------------------------------------------------------------------------------
        {
            SIOMM_DigBankReadArea Data = new SIOMM_DigBankReadArea();

            // Read the data
            byte[] arrbyData = ReadBlock(SIOMM.MM.DBANK.READ_AREA_BASE, 32);

            // If everything is okay, go ahead and fill the structure
            Data.StatePoints = O22SnapIoHelper.ToUInt64(arrbyData, 0);
            Data.OnLatchStatePoints = O22SnapIoHelper.ToUInt64(arrbyData, 8);
            Data.OffLatchStatePoints = O22SnapIoHelper.ToUInt64(arrbyData, 16);
            Data.ActiveCountersPoints = O22SnapIoHelper.ToUInt64(arrbyData, 24);

            return Data;
        }


        public UInt32 ReadClearDigPtCounts(short Point)
        //-------------------------------------------------------------------------------------------------
        // Read and clear counts for the specified digital point
        //-------------------------------------------------------------------------------------------------
        {
            return ReadQuad(SIOMM.MM.POINT.DPOINT.READ_CLEAR_COUNTS_BASE + (UInt32)(SIOMM.MM.POINT.DPOINT.READ_CLEAR_BOUNDARY * Point));
        }

        public UInt32 ReadClearDigPtCountsX(short Point)
        //-------------------------------------------------------------------------------------------------
        // Read and clear counts for the specified digital point
        //-------------------------------------------------------------------------------------------------
        {
            return ReadQuad(SIOMM.MM.POINT.DPOINT.READ_CLEAR_COUNTS_BASE_X + (UInt32)(SIOMM.MM.POINT.DPOINT.READ_CLEAR_BOUNDARY_X * Point));
        }

        public bool ReadClearDigPtOnLatch(short Point)
        //-------------------------------------------------------------------------------------------------
        // Read and clear the on-latch state for the specified digital point
        //-------------------------------------------------------------------------------------------------
        {
            UInt32 quad = ReadQuad(SIOMM.MM.POINT.DPOINT.READ_CLEAR_ON_LATCH_BASE + (UInt32)(SIOMM.MM.POINT.DPOINT.READ_CLEAR_BOUNDARY * Point));           
            return (quad != 0);
        }

        public bool ReadClearDigPtOnLatchX(short Point)
        //-------------------------------------------------------------------------------------------------
        // Read and clear the on-latch state for the specified digital point
        //-------------------------------------------------------------------------------------------------
        {
            UInt32 quad = ReadQuad(SIOMM.MM.POINT.DPOINT.READ_CLEAR_ON_LATCH_BASE_X + (UInt32)(SIOMM.MM.POINT.DPOINT.READ_CLEAR_BOUNDARY_X * Point));
            return (quad != 0);
        }

        public bool ReadClearDigPtOffLatch(short Point)
        //-------------------------------------------------------------------------------------------------
        // Read and clear the off-latch state for the specified digital point
        //-------------------------------------------------------------------------------------------------
        {
            UInt32 quad = ReadQuad(SIOMM.MM.POINT.DPOINT.READ_CLEAR_OFF_LATCH_BASE + (UInt32)(SIOMM.MM.POINT.DPOINT.READ_CLEAR_BOUNDARY * Point));
            return (quad != 0);
        }

        public bool ReadClearDigPtOffLatchX(short Point)
        //-------------------------------------------------------------------------------------------------
        // Read and clear the off-latch state for the specified digital point
        //-------------------------------------------------------------------------------------------------
        {
            UInt32 quad = ReadQuad(SIOMM.MM.POINT.DPOINT.READ_CLEAR_OFF_LATCH_BASE_X + (UInt32)(SIOMM.MM.POINT.DPOINT.READ_CLEAR_BOUNDARY_X * Point));
            return (quad != 0);
        }

        public Single GetAnaPtValue(short Point)
        //-------------------------------------------------------------------------------------------------
        // Get the value of the specified analog point.
        //-------------------------------------------------------------------------------------------------
        {
            return ReadFloat(SIOMM.MM.POINT.APOINT.READ_VALUE_BASE + (UInt32)(SIOMM.MM.POINT.APOINT.READ_BOUNDARY * Point));
        }

        public Single GetAnaPtValueX(short Point)
        //-------------------------------------------------------------------------------------------------
        // Get the value of the specified analog point.
        //-------------------------------------------------------------------------------------------------
        {
            return ReadFloat(SIOMM.MM.POINT.APOINT.READ_VALUE_BASE_X + (UInt32)(SIOMM.MM.POINT.APOINT.READ_BOUNDARY_X * Point));
        }

        public Single GetAnaPtCounts(short Point)
        //-------------------------------------------------------------------------------------------------
        // Get the raw counts of the specified analog point.
        //-------------------------------------------------------------------------------------------------
        {
            return ReadFloat(SIOMM.MM.POINT.APOINT.READ_COUNTS_BASE + (UInt32)(SIOMM.MM.POINT.APOINT.READ_BOUNDARY * Point));
        }

        public Single GetAnaPtCountsX(short Point)
        //-------------------------------------------------------------------------------------------------
        // Get the raw counts of the specified analog point.
        //-------------------------------------------------------------------------------------------------
        {
            return ReadFloat(SIOMM.MM.POINT.APOINT.READ_COUNTS_BASE_X + (UInt32)(SIOMM.MM.POINT.APOINT.READ_BOUNDARY_X * Point));
        }

        public Single GetAnaPtMinValue(short Point)
        //-------------------------------------------------------------------------------------------------
        // Get the minimum value of the specified analog point.
        //-------------------------------------------------------------------------------------------------
        {
            return ReadFloat(SIOMM.MM.POINT.APOINT.READ_MIN_VALUE_BASE + (UInt32)(SIOMM.MM.POINT.APOINT.READ_BOUNDARY * Point));
        }

        public Single GetAnaPtMinValueX(short Point)
        //-------------------------------------------------------------------------------------------------
        // Get the minimum value of the specified analog point.
        //-------------------------------------------------------------------------------------------------
        {
            return ReadFloat(SIOMM.MM.POINT.APOINT.READ_MIN_VALUE_BASE_X + (UInt32)(SIOMM.MM.POINT.APOINT.READ_BOUNDARY_X * Point));
        }

        public Single GetAnaPtMaxValue(short Point)
        //-------------------------------------------------------------------------------------------------
        // Get the maximum value of the specified analog point.
        //-------------------------------------------------------------------------------------------------
        {
            return ReadFloat(SIOMM.MM.POINT.APOINT.READ_MAX_VALUE_BASE + (UInt32)(SIOMM.MM.POINT.APOINT.READ_BOUNDARY * Point));
        }

        public Single GetAnaPtMaxValueX(short Point)
        //-------------------------------------------------------------------------------------------------
        // Get the maximum value of the specified analog point.
        //-------------------------------------------------------------------------------------------------
        {
            return ReadFloat(SIOMM.MM.POINT.APOINT.READ_MAX_VALUE_BASE_X + (UInt32)(SIOMM.MM.POINT.APOINT.READ_BOUNDARY_X * Point));
        }

        public Single GetAnaPtTpoPeriod(short Point)
        //-------------------------------------------------------------------------------------------------
        // Get the TPO period of the specified analog point.
        //-------------------------------------------------------------------------------------------------
        {
            return ReadFloat(SIOMM.MM.POINT.APOINT.READ_TPO_PERIOD_BASE + (UInt32)(SIOMM.MM.POINT.APOINT.READ_BOUNDARY * Point));
        }

        public Single GetAnaPtTpoPeriodX(short Point)
        //-------------------------------------------------------------------------------------------------
        // Get the TPO period of the specified analog point.
        //-------------------------------------------------------------------------------------------------
        {
            return ReadFloat(SIOMM.MM.POINT.APOINT.READ_TPO_PERIOD_BASE_X + (UInt32)(SIOMM.MM.POINT.APOINT.READ_BOUNDARY_X * Point));
        }

        public SIOMM_AnaPointReadArea GetAnaPtReadAreaEx(short Point)
        //-------------------------------------------------------------------------------------------------
        // Get the read area for the specified analog point
        //-------------------------------------------------------------------------------------------------
        {
            SIOMM_AnaPointReadArea Data = new SIOMM_AnaPointReadArea();

            // Read the data
            byte[] arrbyData = ReadBlock(SIOMM.MM.POINT.APOINT.READ_AREA_BASE + (UInt32)(SIOMM.MM.POINT.APOINT.READ_BOUNDARY * Point), 16);

            // If everything is okay, go ahead and fill the structure
            Data.Value = O22SnapIoHelper.ToSingle(arrbyData, 0);
            Data.Counts = O22SnapIoHelper.ToSingle(arrbyData, 4);
            Data.MinValue = O22SnapIoHelper.ToSingle(arrbyData, 8);
            Data.MaxValue = O22SnapIoHelper.ToSingle(arrbyData, 12);

            return Data;
        }

        public SIOMM_AnaPointReadArea GetAnaPtReadAreaExX(short Point)
        //-------------------------------------------------------------------------------------------------
        // Get the read area for the specified analog point
        //-------------------------------------------------------------------------------------------------
        {
            SIOMM_AnaPointReadArea Data = new SIOMM_AnaPointReadArea();

            // Read the data
            byte[] arrbyData = ReadBlock(SIOMM.MM.POINT.APOINT.READ_AREA_BASE_X + (UInt32)(SIOMM.MM.POINT.APOINT.READ_BOUNDARY_X * Point), 16);

            // If everything is okay, go ahead and fill the structure
            Data.Value = O22SnapIoHelper.ToSingle(arrbyData, 0);
            Data.Counts = O22SnapIoHelper.ToSingle(arrbyData, 4);
            Data.MinValue = O22SnapIoHelper.ToSingle(arrbyData, 8);
            Data.MaxValue = O22SnapIoHelper.ToSingle(arrbyData, 12);

            return Data;
        }

        public void SetAnaPtValue(short Point, Single Value)
        //-------------------------------------------------------------------------------------------------
        // Set the value for the specified analog point
        //-------------------------------------------------------------------------------------------------
        {
            WriteFloat(SIOMM.MM.POINT.APOINT.WRITE_VALUE_BASE + (UInt32)(SIOMM.MM.POINT.APOINT.WRITE_BOUNDARY * Point), Value);
        }

        public void SetAnaPtValueX(short Point, Single Value)
        //-------------------------------------------------------------------------------------------------
        // Set the value for the specified analog point
        //-------------------------------------------------------------------------------------------------
        {
            WriteFloat(SIOMM.MM.POINT.APOINT.WRITE_VALUE_BASE_X + (UInt32)(SIOMM.MM.POINT.APOINT.WRITE_BOUNDARY_X * Point), Value);
        }

        public void SetAnaPtCounts(short Point, Single Value)
        //-------------------------------------------------------------------------------------------------
        // Set the raw counts for the specified analog point
        //-------------------------------------------------------------------------------------------------
        {
            WriteFloat(SIOMM.MM.POINT.APOINT.WRITE_COUNTS_BASE + (UInt32)(SIOMM.MM.POINT.APOINT.WRITE_BOUNDARY * Point), Value);
        }

        public void SetAnaPtCountsX(short Point, Single Value)
        //-------------------------------------------------------------------------------------------------
        // Set the raw counts for the specified analog point
        //-------------------------------------------------------------------------------------------------
        {
            WriteFloat(SIOMM.MM.POINT.APOINT.WRITE_COUNTS_BASE_X + (UInt32)(SIOMM.MM.POINT.APOINT.WRITE_BOUNDARY_X * Point), Value);
        }

        public void SetAnaPtTpoPeriod(short Point, Single Value)
        //-------------------------------------------------------------------------------------------------
        // Set the TPO period for the specified analog point
        //-------------------------------------------------------------------------------------------------
        {
            WriteFloat(SIOMM.MM.POINT.APOINT.WRITE_TPO_PERIOD_BASE + (UInt32)(SIOMM.MM.POINT.APOINT.WRITE_BOUNDARY * Point), Value);
        }

        public void SetAnaPtTpoPeriodX(short Point, Single Value)
        //-------------------------------------------------------------------------------------------------
        // Set the TPO period for the specified analog point
        //-------------------------------------------------------------------------------------------------
        {
            WriteFloat(SIOMM.MM.POINT.APOINT.WRITE_TPO_PERIOD_BASE_X + (UInt32)(SIOMM.MM.POINT.APOINT.WRITE_BOUNDARY_X * Point), Value);
        }

        public Single ReadClearAnaPtMinValue(short Point)
        //-------------------------------------------------------------------------------------------------
        // Read and clear the minimum value for the specified analog point
        //-------------------------------------------------------------------------------------------------
        {
            return ReadFloat(SIOMM.MM.POINT.APOINT.READ_CLEAR_MIN_VALUE_BASE + (UInt32)(SIOMM.MM.POINT.APOINT.READ_CLEAR_BOUNDARY * Point));
        }

        public Single ReadClearAnaPtMinValueX(short Point)
        //-------------------------------------------------------------------------------------------------
        // Read and clear the minimum value for the specified analog point
        //-------------------------------------------------------------------------------------------------
        {
            return ReadFloat(SIOMM.MM.POINT.APOINT.READ_CLEAR_MIN_VALUE_BASE_X + (UInt32)(SIOMM.MM.POINT.APOINT.READ_CLEAR_BOUNDARY_X * Point));
        }

        public Single ReadClearAnaPtMaxValue(short Point)
        //-------------------------------------------------------------------------------------------------
        // Read and clear the maximum value for the specified analog point
        //-------------------------------------------------------------------------------------------------
        {
            return ReadFloat(SIOMM.MM.POINT.APOINT.READ_CLEAR_MAX_VALUE_BASE + (UInt32)(SIOMM.MM.POINT.APOINT.READ_CLEAR_BOUNDARY * Point));
        }

        public Single ReadClearAnaPtMaxValueX(short Point)
        //-------------------------------------------------------------------------------------------------
        // Read and clear the maximum value for the specified analog point
        //-------------------------------------------------------------------------------------------------
        {
            return ReadFloat(SIOMM.MM.POINT.APOINT.READ_CLEAR_MAX_VALUE_BASE_X + (UInt32)(SIOMM.MM.POINT.APOINT.READ_CLEAR_BOUNDARY_X * Point));
        }

        public void ConfigurePoint(short Point, UInt32 PointType)
        //-------------------------------------------------------------------------------------------------
        // Configure the specified point
        //-------------------------------------------------------------------------------------------------
        {
            WriteQuad(SIOMM.MM.POINT.CONFIG_WRITE_TYPE_BASE + (UInt32)(SIOMM.MM.POINT.CONFIG_BOUNDARY * Point), PointType);
        }

        public void ConfigurePointX(short Point, UInt32 PointType)
        //-------------------------------------------------------------------------------------------------
        // Configure the specified point
        //-------------------------------------------------------------------------------------------------
        {
            WriteQuad(SIOMM.MM.POINT.CONFIG_WRITE_TYPE_BASE_X + (UInt32)(SIOMM.MM.POINT.CONFIG_BOUNDARY_X * Point), PointType);
        }

        public UInt32 GetModuleType(short Point)
        //-------------------------------------------------------------------------------------------------
        // Get the module type at the specified point position
        //-------------------------------------------------------------------------------------------------
        {
            return ReadQuad(SIOMM.MM.POINT.CONFIG_READ_MOD_TYPE_BASE + (UInt32)(SIOMM.MM.POINT.CONFIG_BOUNDARY * Point));
        }

        public UInt32 GetModuleTypeX(short Point)
        //-------------------------------------------------------------------------------------------------
        // Get the module type at the specified point position
        //-------------------------------------------------------------------------------------------------
        {
            return ReadQuad(SIOMM.MM.POINT.CONFIG_READ_MOD_TYPE_BASE_X + (UInt32)(SIOMM.MM.POINT.CONFIG_BOUNDARY_X * Point));
        }

        public void SetDigPtConfiguration(short Point, UInt32 PointType, UInt32 Feature)
        //-------------------------------------------------------------------------------------------------
        // Configure a digital point
        //-------------------------------------------------------------------------------------------------
        {
            byte[] arrbyData = new byte[8]; // buffer for the data to be written

            // Build the data area
            O22SnapIoHelper.GetBytes(PointType).CopyTo(arrbyData, 0);
            O22SnapIoHelper.GetBytes(Feature).CopyTo(arrbyData, 4);

            // Write the data
            WriteBlock(SIOMM.MM.POINT.CONFIG_WRITE_TYPE_BASE + (UInt32)(SIOMM.MM.POINT.CONFIG_BOUNDARY * Point), arrbyData);
        }

        public void SetDigPtConfigurationX(short Point, UInt32 PointType, UInt32 Feature)
        //-------------------------------------------------------------------------------------------------
        // Configure a digital point
        //-------------------------------------------------------------------------------------------------
        {
            byte[] arrbyData = new byte[8]; // buffer for the data to be written

            // Build the data area
            O22SnapIoHelper.GetBytes(PointType).CopyTo(arrbyData, 0);
            O22SnapIoHelper.GetBytes(Feature).CopyTo(arrbyData, 4);

            // Write the data
            WriteBlock(SIOMM.MM.POINT.CONFIG_WRITE_TYPE_BASE_X + (UInt32)(SIOMM.MM.POINT.CONFIG_BOUNDARY_X * Point), arrbyData);
        }

       public void SetAnaPtConfiguration(short Point, UInt32 PointType,
                                                   Single Offset, Single Gain,
                                                   Single HiScale, Single LoScale)
       //-------------------------------------------------------------------------------------------------
       // Configure an analog point
       //-------------------------------------------------------------------------------------------------
       {
           byte[] arrbyData = new byte[24]; // buffer for the data to be read

           // Build the data area
           O22SnapIoHelper.GetBytes(PointType).CopyTo(arrbyData, 0);
           O22SnapIoHelper.GetBytes(0).CopyTo(arrbyData, 4);
           O22SnapIoHelper.GetBytes(Offset).CopyTo(arrbyData, 8);
           O22SnapIoHelper.GetBytes(Gain).CopyTo(arrbyData, 12);
           O22SnapIoHelper.GetBytes(HiScale).CopyTo(arrbyData, 16);
           O22SnapIoHelper.GetBytes(LoScale).CopyTo(arrbyData, 20);

           // Write the data
           WriteBlock(SIOMM.MM.POINT.CONFIG_WRITE_TYPE_BASE + (UInt32)(SIOMM.MM.POINT.CONFIG_BOUNDARY * Point), arrbyData);
       }

       public void SetAnaPtConfigurationX(short Point, UInt32 PointType,
                                           Single Offset, Single Gain,
                                           Single HiScale, Single LoScale)
       //-------------------------------------------------------------------------------------------------
       // Configure an analog point
       //-------------------------------------------------------------------------------------------------
       {
           byte[] arrbyData = new byte[24]; // buffer for the data to be read

           // Build the data area
           O22SnapIoHelper.GetBytes(PointType).CopyTo(arrbyData, 0);
           O22SnapIoHelper.GetBytes(0).CopyTo(arrbyData, 4);
           O22SnapIoHelper.GetBytes(Offset).CopyTo(arrbyData, 8);
           O22SnapIoHelper.GetBytes(Gain).CopyTo(arrbyData, 12);
           O22SnapIoHelper.GetBytes(HiScale).CopyTo(arrbyData, 16);
           O22SnapIoHelper.GetBytes(LoScale).CopyTo(arrbyData, 20);  //TODO: should the parameter be a struct, should the clamp values be added.

           // Write the data
           WriteBlock(SIOMM.MM.POINT.CONFIG_WRITE_TYPE_BASE_X + (UInt32)(SIOMM.MM.POINT.CONFIG_BOUNDARY_X * Point), arrbyData);
       }

       public void SetPtConfigurationEx2X(short Point, SIOMM_PointConfigArea2X PtConfigData)
       //-------------------------------------------------------------------------------------------------
       // Configure a digital or analog point
       //-------------------------------------------------------------------------------------------------
       {
           byte[] arrbyData = new byte[SIOMM.MM.POINT.CONFIG_BOUNDARY_X-4]; // buffer for the data to be written

           // Build the data area - skip Module Type as it is read only
           O22SnapIoHelper.GetBytes(PtConfigData.PointType).CopyTo(arrbyData, 0);
           O22SnapIoHelper.GetBytes(PtConfigData.Feature).CopyTo(arrbyData, 4);
           O22SnapIoHelper.GetBytes(PtConfigData.Offset).CopyTo(arrbyData, 8);
           O22SnapIoHelper.GetBytes(PtConfigData.Gain).CopyTo(arrbyData, 12);
           O22SnapIoHelper.GetBytes(PtConfigData.HiScale).CopyTo(arrbyData, 16);
           O22SnapIoHelper.GetBytes(PtConfigData.LoScale).CopyTo(arrbyData, 20);
           // bytes 24-27 are not used at this time. 
           O22SnapIoHelper.GetBytes(PtConfigData.FilterWeight).CopyTo(arrbyData, 28);
           O22SnapIoHelper.GetBytes(PtConfigData.WatchdogValue).CopyTo(arrbyData, 32);
           O22SnapIoHelper.GetBytes(PtConfigData.WatchdogEnabled).CopyTo(arrbyData, 39);  //36-38 are zero
           //bytes 40-43 are not used at this time.
           //PtConfigData.bName.CopyTo(arrbyData, 44);

            Encoding.ASCII.GetBytes(PtConfigData.Name, 0, (int)SIOMM.MM.POINT.CONFIG_NAME_SIZE_X, arrbyData, 44);

           //bytes 95-179 are reserved
           O22SnapIoHelper.GetBytes(PtConfigData.LoClamp).CopyTo(arrbyData, 180);
           O22SnapIoHelper.GetBytes(PtConfigData.HiClamp).CopyTo(arrbyData, 184);


           // Write the first section
           byte[] Data1 = new byte[24];
           Array.Copy(arrbyData, 0, Data1, 0, 24);
           WriteBlock(SIOMM.MM.POINT.CONFIG_WRITE_TYPE_BASE_X + (UInt32)(SIOMM.MM.POINT.CONFIG_BOUNDARY_X * Point), Data1);

           // Write the second section
           byte[] Data2 = new byte[12];
           Array.Copy(arrbyData, 28, Data2, 0, 12);
           WriteBlock(SIOMM.MM.POINT.CONFIG_WRITE_TYPE_BASE_X + 28 + (UInt32)(SIOMM.MM.POINT.CONFIG_BOUNDARY_X * Point), Data2);

           // Write the third section
           byte[] Data3 = new byte[SIOMM.MM.POINT.CONFIG_NAME_SIZE_X];
           Array.Copy(arrbyData, 44, Data2, 0, SIOMM.MM.POINT.CONFIG_NAME_SIZE_X);
           WriteBlock(SIOMM.MM.POINT.CONFIG_WRITE_TYPE_BASE_X + 44 + (UInt32)(SIOMM.MM.POINT.CONFIG_BOUNDARY_X * Point), Data3);

           // Write the fourth section
           byte[] Data4 = new byte[8];
           Array.Copy(arrbyData, 180, Data4, 0, 8);
           WriteBlock(SIOMM.MM.POINT.CONFIG_WRITE_TYPE_BASE_X + 180 + (UInt32)(SIOMM.MM.POINT.CONFIG_BOUNDARY_X * Point), Data4);
       }
  
       public void SetPtConfigurationEx2(short Point, SIOMM_PointConfigArea2 PtConfigData)
       //-------------------------------------------------------------------------------------------------
       // Configure a digital or analog point
       //-------------------------------------------------------------------------------------------------
       {
         byte [] arrbyData = new byte[SIOMM.MM.POINT.CONFIG_BOUNDARY]; // buffer for the data to be read

           // Build the data area
         O22SnapIoHelper.GetBytes(PtConfigData.PointType).CopyTo(arrbyData, 0);
         O22SnapIoHelper.GetBytes(PtConfigData.Feature).CopyTo(arrbyData, 4);
         O22SnapIoHelper.GetBytes(PtConfigData.Offset).CopyTo(arrbyData, 8);
         O22SnapIoHelper.GetBytes(PtConfigData.Gain).CopyTo(arrbyData, 12);
         O22SnapIoHelper.GetBytes(PtConfigData.HiScale).CopyTo(arrbyData, 16);
         O22SnapIoHelper.GetBytes(PtConfigData.LoScale).CopyTo(arrbyData, 20);
         // bytes 24-27 are not used at this time. 
         O22SnapIoHelper.GetBytes(PtConfigData.FilterWeight).CopyTo(arrbyData, 28);
         O22SnapIoHelper.GetBytes(PtConfigData.WatchdogValue).CopyTo(arrbyData, 32);
         O22SnapIoHelper.GetBytes(PtConfigData.WatchdogEnabled).CopyTo(arrbyData, 39);  //36-38 are zero

         //bytes 40-43 are not used at this time.
         PtConfigData.bName.CopyTo(arrbyData, 44);

         // Write the first section
         WriteBlock(SIOMM.MM.POINT.CONFIG_WRITE_TYPE_BASE + (UInt32)(SIOMM.MM.POINT.CONFIG_BOUNDARY * Point), arrbyData);

         // Write the second section
         byte[] Data2 = new byte[12];
         Array.Copy(arrbyData, 28, Data2, 0, 12);
         WriteBlock(SIOMM.MM.POINT.CONFIG_WRITE_TYPE_BASE + 28 + (UInt32)(SIOMM.MM.POINT.CONFIG_BOUNDARY * Point), Data2);

         // Write the third section
         byte[] Data3 = new byte[SIOMM.MM.POINT.CONFIG_NAME_SIZE];
         Array.Copy(arrbyData, 44, Data2, 0, SIOMM.MM.POINT.CONFIG_NAME_SIZE);
         WriteBlock(SIOMM.MM.POINT.CONFIG_WRITE_TYPE_BASE + 44 + (UInt32)(SIOMM.MM.POINT.CONFIG_BOUNDARY * Point), Data3);
       }


       public void SetPtConfigurationEx(short Point, SIOMM_PointConfigArea PtConfigData)
       //-------------------------------------------------------------------------------------------------
       // Configure a digital or analog point
       //-------------------------------------------------------------------------------------------------
       {
           byte[] arrbyData = new byte[40]; // buffer for the data to be read

           // Build the data area
           O22SnapIoHelper.GetBytes(PtConfigData.PointType).CopyTo(arrbyData, 0);
           O22SnapIoHelper.GetBytes(PtConfigData.Feature).CopyTo(arrbyData, 4);
           O22SnapIoHelper.GetBytes(PtConfigData.Offset).CopyTo(arrbyData, 8);
           O22SnapIoHelper.GetBytes(PtConfigData.Gain).CopyTo(arrbyData, 12);
           O22SnapIoHelper.GetBytes(PtConfigData.HiScale).CopyTo(arrbyData, 16);
           O22SnapIoHelper.GetBytes(PtConfigData.LoScale).CopyTo(arrbyData, 20);
           // bytes 24-31 are not used at this time. 
           O22SnapIoHelper.GetBytes(PtConfigData.WatchdogValue).CopyTo(arrbyData, 32);
           O22SnapIoHelper.GetBytes(PtConfigData.WatchdogEnabled).CopyTo(arrbyData, 39);  //36-38 are zero

           // Write the first section
           WriteBlock(SIOMM.MM.POINT.CONFIG_WRITE_TYPE_BASE + (UInt32)(SIOMM.MM.POINT.CONFIG_BOUNDARY * Point), arrbyData);

           // Write the second section
           byte[] Data2 = new byte[8];
           Array.Copy(arrbyData, 32, Data2, 0, 8);
           WriteBlock(SIOMM.MM.POINT.CONFIG_WRITE_TYPE_BASE + 32 + (UInt32)(SIOMM.MM.POINT.CONFIG_BOUNDARY * Point), Data2);
       }

       public SIOMM_PointConfigArea2X GetPtConfigurationEx2X(short Point)
       //-------------------------------------------------------------------------------------------------
       // Get the configuration for a point
       //-------------------------------------------------------------------------------------------------
       {
           // Read the data
           byte[] arrbyData = ReadBlock(SIOMM.MM.POINT.CONFIG_READ_MOD_TYPE_BASE_X +
                               (UInt32)(SIOMM.MM.POINT.CONFIG_BOUNDARY_X * Point),
                               (UInt16)SIOMM.MM.POINT.CONFIG_BOUNDARY_X);

            SIOMM_PointConfigArea2X PtConfigData = PopulatePointData(arrbyData, 0);

           return PtConfigData;
       }

        private SIOMM_PointConfigArea2X PopulatePointData(byte [] data, int start)
        {
            SIOMM_PointConfigArea2X PtConfigData = new SIOMM_PointConfigArea2X();
            PtConfigData.Name = String.Empty;

            // If everything is okay, go ahead and fill the structure
            PtConfigData.ModuleType = O22SnapIoHelper.ToUInt32(data, 0 + start);
            if (PtConfigData.ModuleType == UInt32.MaxValue) //Invalid area
                PtConfigData.Valid = false;
            else
            {
                PtConfigData.Valid = true;

                PtConfigData.PointType = O22SnapIoHelper.ToUInt32(data, 4 + start);
                PtConfigData.Feature = O22SnapIoHelper.ToUInt32(data, 8 + start);
                PtConfigData.Offset = O22SnapIoHelper.ToSingle(data, 12 + start);
                PtConfigData.Gain = O22SnapIoHelper.ToSingle(data, 16 + start);
                PtConfigData.HiScale = O22SnapIoHelper.ToSingle(data, 20 + start);
                PtConfigData.LoScale = O22SnapIoHelper.ToSingle(data, 24 + start);
                // bytes 28-31 are not used at this time
                PtConfigData.FilterWeight = O22SnapIoHelper.ToSingle(data, 32 + start);
                PtConfigData.WatchdogValue = O22SnapIoHelper.ToSingle(data, 36 + start);
                PtConfigData.WatchdogEnabled = O22SnapIoHelper.ToBoolean(data, 43 + start);
                //bytes 44-47 are reserved
                PtConfigData.Name = Encoding.ASCII.GetString(data, 48 + start, (int)SIOMM.MM.POINT.CONFIG_NAME_SIZE_X).TrimEnd('\0');
                //Array.Copy(arrbyData, 48, PtConfigData.bName, 0, (int)SIOMM.MM.POINT.CONFIG_NAME_SIZE_X);

                //bytes 99-183 are reserved
                PtConfigData.LoClamp = O22SnapIoHelper.ToSingle(data, 184 + start);
                PtConfigData.HiClamp = O22SnapIoHelper.ToSingle(data, 188 + start);
            }
            return PtConfigData;

        }

       public SIOMM_PointConfigArea2X [] GetPtConfigurationEx2XBulk(int StartPoint, int Qty)
        {
            if (Qty + StartPoint > 4096)
                Qty = 4096 - StartPoint;

            // Read the data
            byte[] arrbyData = ReadBlock(SIOMM.MM.POINT.CONFIG_READ_MOD_TYPE_BASE_X +
                                (UInt32)(SIOMM.MM.POINT.CONFIG_BOUNDARY_X * StartPoint),
                                (UInt32)(SIOMM.MM.POINT.CONFIG_BOUNDARY_X * Qty));

            SIOMM_PointConfigArea2X[] PtConfigData = new SIOMM_PointConfigArea2X[Qty];

            for (int i=0; i< Qty; i++)
            {
                PtConfigData[i] = PopulatePointData(arrbyData, (int)(i * SIOMM.MM.POINT.CONFIG_BOUNDARY_X));
            }

            return PtConfigData;
        }

       public SIOMM_PointConfigArea2 GetPtConfigurationEx2(short Point)
       //-------------------------------------------------------------------------------------------------
       // Get the configuration for a point
       //-------------------------------------------------------------------------------------------------
       {
           SIOMM_PointConfigArea2 PtConfigData = new SIOMM_PointConfigArea2();

           // Read the data
           byte[] arrbyData = ReadBlock(SIOMM.MM.POINT.CONFIG_READ_MOD_TYPE_BASE +
                               (UInt32)(SIOMM.MM.POINT.CONFIG_BOUNDARY * Point),
                               (UInt16)SIOMM.MM.POINT.CONFIG_BOUNDARY);

           // If everything is okay, go ahead and fill the structure
           PtConfigData.ModuleType = O22SnapIoHelper.ToUInt32(arrbyData, 0);
           PtConfigData.PointType = O22SnapIoHelper.ToUInt32(arrbyData, 4);
           PtConfigData.Feature = O22SnapIoHelper.ToUInt32(arrbyData, 8);
           PtConfigData.Offset = O22SnapIoHelper.ToSingle(arrbyData, 12);
           PtConfigData.Gain = O22SnapIoHelper.ToSingle(arrbyData, 16);
           PtConfigData.HiScale = O22SnapIoHelper.ToSingle(arrbyData, 20);
           PtConfigData.LoScale = O22SnapIoHelper.ToSingle(arrbyData, 24);
           // bytes 28-31 are not used at this time
           PtConfigData.FilterWeight = O22SnapIoHelper.ToSingle(arrbyData, 32);
           PtConfigData.WatchdogValue = O22SnapIoHelper.ToSingle(arrbyData, 36);
           PtConfigData.WatchdogEnabled = O22SnapIoHelper.ToBoolean(arrbyData, 43);

           Array.Copy(arrbyData, 48, PtConfigData.bName, 0, (int)SIOMM.MM.POINT.CONFIG_NAME_SIZE);

           return PtConfigData;
       }

       public SIOMM_PointConfigArea GetPtConfigurationEx(short Point)
       //-------------------------------------------------------------------------------------------------
       // Get the configuration for a point
       //-------------------------------------------------------------------------------------------------
       {
           SIOMM_PointConfigArea PtConfigData = new SIOMM_PointConfigArea();

           // Read the data
           byte[] arrbyData = ReadBlock(SIOMM.MM.POINT.CONFIG_READ_MOD_TYPE_BASE + (UInt32)(SIOMM.MM.POINT.CONFIG_BOUNDARY * Point), 44);

           // If everything is okay, go ahead and fill the structure
           PtConfigData.ModuleType = O22SnapIoHelper.ToUInt32(arrbyData, 0);
           PtConfigData.PointType = O22SnapIoHelper.ToUInt32(arrbyData, 4);
           PtConfigData.Feature = O22SnapIoHelper.ToUInt32(arrbyData, 8);
           PtConfigData.Offset = O22SnapIoHelper.ToSingle(arrbyData, 12);
           PtConfigData.Gain = O22SnapIoHelper.ToSingle(arrbyData, 16);
           PtConfigData.HiScale = O22SnapIoHelper.ToSingle(arrbyData, 20);
           PtConfigData.LoScale = O22SnapIoHelper.ToSingle(arrbyData, 24);
           // bytes 28-35 are not used at this time
           PtConfigData.WatchdogValue = O22SnapIoHelper.ToSingle(arrbyData, 36);
           PtConfigData.WatchdogEnabled = O22SnapIoHelper.ToBoolean(arrbyData, 43);

           return PtConfigData;
       }

       public void SetPtWatchdog(short Point, Single Value, bool Enabled)
       //-------------------------------------------------------------------------------------------------
       // Set the watchdog values for a point.
       //-------------------------------------------------------------------------------------------------
       {
           byte[] arrbyData = new byte[8]; // buffer for the data to be read

           // Build the data area
           O22SnapIoHelper.GetBytes(Value).CopyTo(arrbyData, 0);
           O22SnapIoHelper.GetBytes(Enabled).CopyTo(arrbyData, 7);

           // Write the data
           WriteBlock(SIOMM.MM.POINT.CONFIG_WRITE_WDOG_VALUE_BASE + (UInt32)(SIOMM.MM.POINT.CONFIG_BOUNDARY * Point), arrbyData);
       }

       public void SetPtWatchdogX(short Point, Single Value, bool Enabled)
       //-------------------------------------------------------------------------------------------------
       // Set the watchdog values for a point.
       //-------------------------------------------------------------------------------------------------
       {
           byte[] arrbyData = new byte[8]; // buffer for the data to be read

           // Build the data area
           O22SnapIoHelper.GetBytes(Value).CopyTo(arrbyData, 0);
           O22SnapIoHelper.GetBytes(Enabled).CopyTo(arrbyData, 7);

           // Write the data
           WriteBlock(SIOMM.MM.POINT.CONFIG_WRITE_WDOG_VALUE_BASE_X + (UInt32)(SIOMM.MM.POINT.CONFIG_BOUNDARY_X * Point), arrbyData);
       }

       public Single CalcSetAnaPtOffset(short Point)
       //-------------------------------------------------------------------------------------------------
       // Calculate and set an analog point's offset
       //-------------------------------------------------------------------------------------------------
       {
           return ReadFloat(SIOMM.MM.POINT.APOINT.READ_CALC_SET_OFFSET_BASE +
                            (UInt32)(SIOMM.MM.POINT.APOINT.READ_CALC_SET_BOUNDARY * Point));
       }


       public Single CalcSetAnaPtOffsetX(short Point)
       //-------------------------------------------------------------------------------------------------
       // Calculate and set an analog point's offset
       //-------------------------------------------------------------------------------------------------
       {
           return ReadFloat(SIOMM.MM.POINT.APOINT.READ_CALC_SET_OFFSET_BASE_X +
                            (UInt32)(SIOMM.MM.POINT.APOINT.READ_CALC_SET_BOUNDARY_X * Point));
       }

       public Single CalcSetAnaPtGain(short Point)
       //-------------------------------------------------------------------------------------------------
       // Calculate and set an analog point's gain
       //-------------------------------------------------------------------------------------------------
       {
           return ReadFloat(SIOMM.MM.POINT.APOINT.READ_CALC_SET_GAIN_BASE +
                          (UInt32)(SIOMM.MM.POINT.APOINT.READ_CALC_SET_BOUNDARY * Point));
       }

       public Single CalcSetAnaPtGainX(short Point)
       //-------------------------------------------------------------------------------------------------
       // Calculate and set an analog point's gain
       //-------------------------------------------------------------------------------------------------
       {
           return ReadFloat(SIOMM.MM.POINT.APOINT.READ_CALC_SET_GAIN_BASE_X +
                          (UInt32)(SIOMM.MM.POINT.APOINT.READ_CALC_SET_BOUNDARY_X * Point));
       }

       private bool ReadBool(UInt32 DestOffset)
       {
           UInt32 response = ReadQuad(DestOffset);

           return (response != 0);
       }

       private int GetStatusLastError()
       {
           return (int)ReadQuad(SIOMM.MM.STATUS.READ_LAST_ERROR);
       }

       public bool GetStatusPUC()
       //-------------------------------------------------------------------------------------------------
       // Get the Powerup Clear flag
       //-------------------------------------------------------------------------------------------------
       {
           return ReadBool(SIOMM.MM.STATUS.READ_PUC_FLAG);
       }

       public bool GetStatusBootpAlways()
       //-------------------------------------------------------------------------------------------------
       // Get the "bootp always" flag
       //-------------------------------------------------------------------------------------------------
       {
           return ReadBool(SIOMM.MM.STATUS.READ_BOOTP_FLAG);
       }

       public TemperatureUnits GetStatusDegrees()
       //-------------------------------------------------------------------------------------------------
       // Get the temperature unit. F is true, C is false.
       //-------------------------------------------------------------------------------------------------
       {
           if (ReadBool(SIOMM.MM.STATUS.READ_DEGREES_FLAG))
               return TemperatureUnits.Fahrenheit;
           else
               return TemperatureUnits.Celcius;
       }


       public UInt32 GetStatusWatchdogTime()
       //-------------------------------------------------------------------------------------------------
       // Get the watchdog time
       //-------------------------------------------------------------------------------------------------
       {
           return ReadQuad(SIOMM.MM.STATUS.READ_WATCHDOG_TIME);
       }


       public SIOMM_StatusVersion GetStatusVersionEx()
       //-------------------------------------------------------------------------------------------------
       // Get version information
       //-------------------------------------------------------------------------------------------------
       {
           SIOMM_StatusVersion VersionData = new SIOMM_StatusVersion();

           // Read the data
           byte[] arrbyData = ReadBlock(SIOMM.MM.STATUS.READ_BASE, 32);

           // If everything is okay, go ahead and fill the structure
           VersionData.MapVer = O22SnapIoHelper.ToUInt32(arrbyData, 0);
           VersionData.LoaderVersion.Major = arrbyData[24];
           VersionData.LoaderVersion.Minor = arrbyData[25];
           VersionData.LoaderVersion.Revision = arrbyData[26];
           VersionData.LoaderVersion.Letter = arrbyData[27];
           VersionData.KernelVersion.Major = arrbyData[28];
           VersionData.KernelVersion.Minor = arrbyData[29];
           VersionData.KernelVersion.Revision = arrbyData[30];
           VersionData.KernelVersion.Letter = arrbyData[31];

           return VersionData;
       }

       public SIOMM_StatusHardware GetStatusHardwareEx()
       //-------------------------------------------------------------------------------------------------
       // Get hardware information
       //-------------------------------------------------------------------------------------------------
       {
           SIOMM_StatusHardware HardwareData = new SIOMM_StatusHardware();
           // Read the data
           byte[] arrbyData = ReadBlock(SIOMM.MM.STATUS.READ_BASE, 44);

           HardwareData.IoUnitType = O22SnapIoHelper.ToUInt32(arrbyData, 32);
           HardwareData.HwdVerMonth = arrbyData[36];
           HardwareData.HwdVerDay = arrbyData[37];
           HardwareData.HwdVerYear = O22SnapIoHelper.ToUInt16(arrbyData, 38);
           HardwareData.RamSize = O22SnapIoHelper.ToUInt32(arrbyData, 40);

           return HardwareData;
       }


       public SIOMM_StatusHardware2 GetStatusHardwareEx2()
       //-------------------------------------------------------------------------------------------------
       // Get hardware information
       //-------------------------------------------------------------------------------------------------
       {
           SIOMM_StatusHardware2 HardwareData = new SIOMM_StatusHardware2();

           // Use the original function first
           SIOMM_StatusHardware HardwareData1 = GetStatusHardwareEx();

           // Read the part number
           byte[] arrbyData = ReadBlock(SIOMM.MM.STATUS.READ_PART_NUMBER, (UInt16)SIOMM.MM.STATUS.READ_PART_NUMBER_SIZE);

           // Copy the part number string
           //memcpy(pHardwareData->byPartNumber, arrbyData, SIOMM.MM.STATUS.READ_PART_NUMBER_SIZE);
           arrbyData.CopyTo(HardwareData.bPartNumber, 0);
           // If everything is okay, go ahead and fill the structure        
           HardwareData.IoUnitType = HardwareData1.IoUnitType;
           HardwareData.HwdVerMonth = HardwareData1.HwdVerMonth;
           HardwareData.HwdVerDay = HardwareData1.HwdVerDay;
           HardwareData.HwdVerYear = HardwareData1.HwdVerYear;
           HardwareData.RamSize = HardwareData1.RamSize;

           return HardwareData;
       }


       public SIOMM_StatusNetwork GetStatusNetworkEx()
       //-------------------------------------------------------------------------------------------------
       // Get network information
       //-------------------------------------------------------------------------------------------------
       {
           SIOMM_StatusNetwork NetworkData = new SIOMM_StatusNetwork();

           // Read the data
           byte[] arrbyData = ReadBlock(SIOMM.MM.STATUS.READ_BASE, 64);

           // If everything is okay, go ahead and fill the structure
           byte[] MACConversion = new byte[64];
           Array.Copy(arrbyData, 46, MACConversion, 16, 48);
           NetworkData.MACAddress = O22SnapIoHelper.ToUInt64(MACConversion, 0);
           NetworkData.TCPIPAddress.Octet1 = arrbyData[52];
           NetworkData.TCPIPAddress.Octet2 = arrbyData[53];
           NetworkData.TCPIPAddress.Octet3 = arrbyData[54];
           NetworkData.TCPIPAddress.Octet4 = arrbyData[55];
           NetworkData.SubnetMask.Octet1 = arrbyData[56];
           NetworkData.SubnetMask.Octet2 = arrbyData[57];
           NetworkData.SubnetMask.Octet3 = arrbyData[58];
           NetworkData.SubnetMask.Octet4 = arrbyData[59];
           NetworkData.DefGateway.Octet1 = arrbyData[60];
           NetworkData.DefGateway.Octet2 = arrbyData[61];
           NetworkData.DefGateway.Octet3 = arrbyData[62];
           NetworkData.DefGateway.Octet4 = arrbyData[63];

           return NetworkData;
       }


       public SIOMM_StatusNetwork2 GetStatusNetworkEx2()
       //-------------------------------------------------------------------------------------------------
       // Get network information
       //-------------------------------------------------------------------------------------------------
       {
           SIOMM_StatusNetwork2 NetworkData = new SIOMM_StatusNetwork2();
           // Read the data
           byte[] arrbyData = ReadBlock(SIOMM.MM.STATUS.READ_BASE, 124);

           // If everything is okay, go ahead and fill the structure
           byte[] MACConversion = new byte[64];
           Array.Copy(arrbyData, 46, MACConversion, 16, 48);
           NetworkData.MACAddress = O22SnapIoHelper.ToUInt64(MACConversion, 0);
           NetworkData.TCPIPAddress.Octet1 = arrbyData[52];
           NetworkData.TCPIPAddress.Octet2 = arrbyData[53];
           NetworkData.TCPIPAddress.Octet3 = arrbyData[54];
           NetworkData.TCPIPAddress.Octet4 = arrbyData[55];
           NetworkData.SubnetMask.Octet1 = arrbyData[56];
           NetworkData.SubnetMask.Octet2 = arrbyData[57];
           NetworkData.SubnetMask.Octet3 = arrbyData[58];
           NetworkData.SubnetMask.Octet4 = arrbyData[59];
           NetworkData.DefGateway.Octet1 = arrbyData[60];
           NetworkData.DefGateway.Octet2 = arrbyData[61];
           NetworkData.DefGateway.Octet3 = arrbyData[62];
           NetworkData.DefGateway.Octet4 = arrbyData[63];

           NetworkData.TcpIpMinRtoMS = O22SnapIoHelper.ToUInt32(arrbyData, 88);
           NetworkData.InitialRtoMS = O22SnapIoHelper.ToUInt32(arrbyData, 100);
           NetworkData.TcpRetries = O22SnapIoHelper.ToUInt32(arrbyData, 104);
           NetworkData.TcpIdleTimeout = O22SnapIoHelper.ToUInt32(arrbyData, 108);
           NetworkData.EnetLateCol = O22SnapIoHelper.ToUInt32(arrbyData, 112);
           NetworkData.EnetExcessiveCol = O22SnapIoHelper.ToUInt32(arrbyData, 116);
           NetworkData.EnetOtherErrors = O22SnapIoHelper.ToUInt32(arrbyData, 120);

           return NetworkData;
       }

       public void SetStatusOperation(SIOMM_OpCode OpCode)
       //-------------------------------------------------------------------------------------------------
       // Issue an operation code
       //-------------------------------------------------------------------------------------------------
       {
           WriteQuad(SIOMM.MM.STATUS.WRITE_OPERATION, (UInt32)OpCode);
       }


       public void SetStatusBootpRequest(bool Flag)
       //-------------------------------------------------------------------------------------------------
       // Set the "bootp always" flag
       //-------------------------------------------------------------------------------------------------
       {
           WriteQuad(SIOMM.MM.STATUS.WRITE_BOOTP, (UInt32)(Flag ? 1 : 0));
       }


       public void SetStatusDegrees(TemperatureUnits Units)
       //-------------------------------------------------------------------------------------------------
       // Set the temperature unit
       //-------------------------------------------------------------------------------------------------
       {
           WriteQuad(SIOMM.MM.STATUS.WRITE_TEMP_DEGREES, (UInt32)(TemperatureUnits.Fahrenheit==Units ? 1 : 0));
       }


       public void SetStatusWatchdogTime(UInt32 TimeMS)
       //-------------------------------------------------------------------------------------------------
       // Set the watchdog time
       //-------------------------------------------------------------------------------------------------
       {
           WriteQuad(SIOMM.MM.STATUS.WRITE_WATCHDOG_TIME, TimeMS);
       }


       public DateTime GetDateTime()
       //-------------------------------------------------------------------------------------------------
       // Get the brain's date and time -  Format: YYYY-MM-DD HH:MM:SS.00\0
       //-------------------------------------------------------------------------------------------------
       {
           byte[] arrbyData = ReadBlock(SIOMM.MM.DATE_AND_TIME.BASE, (UInt16)SIOMM.MM.DATE_AND_TIME.SIZE);

           // Copy the date and time string.
           System.Text.ASCIIEncoding ae = new ASCIIEncoding();
           int Year = Int32.Parse(ae.GetString(arrbyData, 0, 4));
           int Month = Int32.Parse(ae.GetString(arrbyData, 5, 2));
           int Day = Int32.Parse(ae.GetString(arrbyData, 8, 2));
           int Hour = Int32.Parse(ae.GetString(arrbyData, 11, 2));
           int Minute = Int32.Parse(ae.GetString(arrbyData, 14, 2));
           int Second = Int32.Parse(ae.GetString(arrbyData, 17, 2));
           int Millisecond = Int32.Parse(ae.GetString(arrbyData, 20, 2));

           return new DateTime(Year, Month, Day, Hour, Minute, Second, Millisecond);
       }


       public void SetDateTime(DateTime DeviceDateTime)
       //-------------------------------------------------------------------------------------------------
       // Set the brain's date and time Format: YYYY-MM-DD HH:MM:SS.00\0
       //-------------------------------------------------------------------------------------------------
       {
           string sDateTime = String.Format("{0:yyyy-MM-dd HH:mm:ss.ff}", DeviceDateTime);
           byte[] bDateTime = new byte[SIOMM.MM.DATE_AND_TIME.SIZE];

           System.Text.ASCIIEncoding ae = new ASCIIEncoding();
           ae.GetBytes(sDateTime).CopyTo(bDateTime, 0);

           WriteBlock(SIOMM.MM.DATE_AND_TIME.BASE, bDateTime);
       }

       private Single[] GetAnaBank(UInt32 DestOffset)
       //-------------------------------------------------------------------------------------------------
       // Get an analog bank section from a location in the SNAP I/O memory map.
       //-------------------------------------------------------------------------------------------------
       {
           Single[] BankData = new Single[SIOMM.MM.ABANK.MAX_ELEMENTS];

           // Read the data
           byte[] arrbyData = ReadBlock(DestOffset, (UInt16)SIOMM.MM.ABANK.MAX_BYTES);

           // Unpack the data packet
           for (int i = 0; i < BankData.Length; i++)
               BankData[i] = O22SnapIoHelper.ToSingle(arrbyData, i * sizeof(Single));

           return BankData;
       }

       private void SetAnaBank(UInt32 DestOffset, Single[] BankData)
       //-------------------------------------------------------------------------------------------------
       // Set an analog bank section to a location in the SNAP I/O memory map.
       //-------------------------------------------------------------------------------------------------
       {
           byte[] arrbyData = new byte[SIOMM.MM.ABANK.MAX_BYTES]; // buffer for the data to be written

           // Pack the data packet
           for (int i = 0; i < SIOMM.MM.ABANK.MAX_ELEMENTS; i++)
               O22SnapIoHelper.GetBytes(BankData[i]).CopyTo(arrbyData, i * sizeof(Single));

           // Write the data
           WriteBlock(DestOffset, arrbyData);
       }

       public Single [] GetAnaBankValuesEx()
       //-------------------------------------------------------------------------------------------------
       // Get the analog point values in the analog bank
       //-------------------------------------------------------------------------------------------------
       {
           return GetAnaBank(SIOMM.MM.ABANK.READ_POINT_VALUES);
       }


       public Single [] GetAnaBankCountsEx()
       //-------------------------------------------------------------------------------------------------
       // Get the analog point raw counts in the analog bank
       //-------------------------------------------------------------------------------------------------
       {
           return GetAnaBank(SIOMM.MM.ABANK.READ_POINT_COUNTS);
       }


       public Single [] GetAnaBankMinValuesEx()
       //-------------------------------------------------------------------------------------------------
       // Get the analog point minimum values in the analog bank
       //-------------------------------------------------------------------------------------------------
       {
           return GetAnaBank(SIOMM.MM.ABANK.READ_POINT_MIN_VALUES);
       }

       public Single [] GetAnaBankMaxValuesEx()
       //-------------------------------------------------------------------------------------------------
       // Get the analog point maximum values in the analog bank
       //-------------------------------------------------------------------------------------------------
       {
           return GetAnaBank(SIOMM.MM.ABANK.READ_POINT_MAX_VALUES);
       }


       public void SetAnaBankValuesEx(Single[] BankData)
       //-------------------------------------------------------------------------------------------------
       // Set the analog point values in the analog bank
       //-------------------------------------------------------------------------------------------------
       {
           SetAnaBank(SIOMM.MM.ABANK.WRITE_POINT_VALUES, BankData);
       }


       public void SetAnaBankCountsEx(Single[] BankData)
       //-------------------------------------------------------------------------------------------------
       // Set the analog point raw counts in the analog bank
       //-------------------------------------------------------------------------------------------------
       {
           SetAnaBank(SIOMM.MM.ABANK.WRITE_POINT_COUNTS, BankData);
       }



       public void SetStreamConfiguration(SIOMM_StreamConfiguration StreamConfig)
       //-------------------------------------------------------------------------------------------------
       // 
       //-------------------------------------------------------------------------------------------------
       {
           byte[] arrbyData = new byte[24]; // buffer for the data to be read

           // Build the data area
           O22SnapIoHelper.GetBytes(StreamConfig.IoMirroringEnabled).CopyTo(arrbyData, 3);
           O22SnapIoHelper.GetBytes(StreamConfig.StartAddress).CopyTo(arrbyData, 4);
           O22SnapIoHelper.GetBytes(StreamConfig.DataSize).CopyTo(arrbyData, 8);
           O22SnapIoHelper.GetBytes(StreamConfig.OnFlag).CopyTo(arrbyData, 15);
           O22SnapIoHelper.GetBytes(StreamConfig.IntervalMS).CopyTo(arrbyData, 16);
           O22SnapIoHelper.GetBytes(StreamConfig.Port).CopyTo(arrbyData, 20);

           // Write the data
           WriteBlock(SIOMM.MM.STREAM.CONFIG_BASE, arrbyData);
       }


       public void SetStreamTarget(int Target, SIOMM_IPAddress IpAddressArg)
       //-------------------------------------------------------------------------------------------------
       // 
       //-------------------------------------------------------------------------------------------------
       {
           UInt32 Temp;

           // Make an integer out of the IP address parts
           Temp = (UInt32)IpAddressArg.Octet1 << 24;
           Temp += (UInt32)IpAddressArg.Octet2 << 16;
           Temp += (UInt32)IpAddressArg.Octet3 << 8;
           Temp += IpAddressArg.Octet4;

           WriteQuad((UInt32)(SIOMM.MM.STREAM.TARGET_BASE + (SIOMM.MM.STREAM.TARGET_BOUNDARY * (Target - 1))), Temp);
       }

       public SIOMM_StreamConfiguration GetStreamConfiguration()
       //-------------------------------------------------------------------------------------------------
       // 
       //-------------------------------------------------------------------------------------------------
       {
           SIOMM_StreamConfiguration StreamConfig = new SIOMM_StreamConfiguration();

           // Read the data
           byte[] arrbyData = ReadBlock(SIOMM.MM.STREAM.CONFIG_BASE, 24);

           // If everything is okay, go ahead and fill the structure
           StreamConfig.IoMirroringEnabled = O22SnapIoHelper.ToBoolean(arrbyData, 3);
           StreamConfig.StartAddress = O22SnapIoHelper.ToUInt32(arrbyData, 4);
           StreamConfig.DataSize = O22SnapIoHelper.ToUInt32(arrbyData, 8);
           StreamConfig.OnFlag = O22SnapIoHelper.ToBoolean(arrbyData, 15);
           StreamConfig.IntervalMS = O22SnapIoHelper.ToUInt32(arrbyData, 16);
           StreamConfig.Port = O22SnapIoHelper.ToUInt32(arrbyData, 20);

           return StreamConfig;
       }


       public SIOMM_IPAddress GetStreamTarget(int Target)
       //-------------------------------------------------------------------------------------------------
       // 
       //-------------------------------------------------------------------------------------------------
       {
           SIOMM_IPAddress IpAddress = new SIOMM_IPAddress();

           // Read the data
           UInt32 temp = ReadQuad((UInt32)(SIOMM.MM.STREAM.TARGET_BASE + (SIOMM.MM.STREAM.TARGET_BOUNDARY * (Target - 1))));

           IpAddress.Octet1 = (byte)((temp & 0xFF000000) >> 24);
           IpAddress.Octet2 = (byte)((temp & 0x00FF0000) >> 16);
           IpAddress.Octet3 = (byte)((temp & 0x0000FF00) >> 8);
           IpAddress.Octet4 = (byte)(temp & 0x000000FF);

           return IpAddress;
       }

       public SIOMM_StreamStandardBlock GetStreamReadAreaEx()
       //-------------------------------------------------------------------------------------------------
       // 
       //-------------------------------------------------------------------------------------------------
       {
           SIOMM_StreamStandardBlock StreamData = new SIOMM_StreamStandardBlock();

           // Read the data
           byte[] arrbyData = ReadBlock(SIOMM.MM.STREAM.READ_AREA_BASE, (UInt16)SIOMM.MM.STREAM.READ_AREA_SIZE);

           StreamData.DataLength = O22SnapIoHelper.ToUInt16(arrbyData, 0);

           UInt16 temp = O22SnapIoHelper.ToUInt16(arrbyData, 2);
           StreamData.Tag = (byte)(temp >> 14);
           StreamData.Channel = (byte)((temp & 0x3F00) >> 8);  //0011 1111 0000 0000
           StreamData.TransactionCode = (byte)((temp & 0x00F0) >> 4);  //0x00F0
           StreamData.SyncCode = (byte)(temp & 0x000F);

           StreamData.AnalogValue = new Single[64];
           for (int i = 0; i < 64; i++)
               StreamData.AnalogValue[i] = O22SnapIoHelper.ToSingle(arrbyData, 4 + (4 * i));

           StreamData.DigPointFeature = new UInt32[64];
           for (int i = 0; i < 64; i++)
               StreamData.DigPointFeature[i] = O22SnapIoHelper.ToUInt32(arrbyData, 4 + (4 * i));

           StreamData.State = O22SnapIoHelper.ToUInt64(arrbyData, 516);
           StreamData.OnLatchState = O22SnapIoHelper.ToUInt64(arrbyData, 524);
           StreamData.OffLatchState = O22SnapIoHelper.ToUInt64(arrbyData, 532);
           StreamData.ActiveCounters = O22SnapIoHelper.ToUInt64(arrbyData, 540);
           //56 bytes reserved
           StreamData.TCPIPAddress.Octet1 = arrbyData[604];
           StreamData.TCPIPAddress.Octet2 = arrbyData[605];
           StreamData.TCPIPAddress.Octet3 = arrbyData[606];
           StreamData.TCPIPAddress.Octet4 = arrbyData[607];

           return StreamData;
       }

       public SIOMM_SerialModuleConfigArea GetSerialModuleConfigurationEx(int SerialPort)
       //-------------------------------------------------------------------------------------------------
       // 
       //-------------------------------------------------------------------------------------------------
       {
           SIOMM_SerialModuleConfigArea ConfigData = new SIOMM_SerialModuleConfigArea();

           // Read the data
           byte[] arrbyData = ReadBlock((UInt32)(SIOMM.MM.SERIAL_MODULE.CONFIG_BASE +
                               (SIOMM.MM.SERIAL_MODULE.CONFIG_BOUNDARY * SerialPort)),
                               (UInt16)SIOMM.MM.SERIAL_MODULE.CONFIG_BOUNDARY);

           UInt32 dwEomData = ReadQuad((UInt32)(SIOMM.MM.SERIAL_MODULE.EOM_BASE +
                                  (SIOMM.MM.SERIAL_MODULE.EOM_BOUNDARY * SerialPort)));

           ConfigData.IpPort = O22SnapIoHelper.ToUInt32(arrbyData, 0);
           ConfigData.BaudRate = O22SnapIoHelper.ToUInt32(arrbyData, 5);
           ConfigData.Parity = arrbyData[8];
           ConfigData.DataBits = arrbyData[9];
           ConfigData.StopBits = arrbyData[10];
           ConfigData.HardwareFlowControl = (arrbyData[11] != 0);
           ConfigData.SendTestMessage = (arrbyData[12] != 0);
           ConfigData.EOM1 = O22SnapIoHelper.GetBytes(dwEomData)[0];
           ConfigData.EOM2 = O22SnapIoHelper.GetBytes(dwEomData)[1];
           ConfigData.EOM3 = O22SnapIoHelper.GetBytes(dwEomData)[2];
           ConfigData.EOM4 = O22SnapIoHelper.GetBytes(dwEomData)[3];

           return ConfigData;
       }

       public void SetSerialModuleConfigurationEx(int SerialPort, SIOMM_SerialModuleConfigArea ConfigData)
       //-------------------------------------------------------------------------------------------------
       // 
       //-------------------------------------------------------------------------------------------------
       {
           byte[] arrbyData = new byte[SIOMM.MM.SERIAL_MODULE.CONFIG_BOUNDARY]; // buffer for the data to be read
           UInt32 dwEomTemp;

           O22SnapIoHelper.GetBytes(ConfigData.BaudRate).CopyTo(arrbyData, 0);
           arrbyData[4] = ConfigData.Parity;
           arrbyData[5] = ConfigData.DataBits;
           arrbyData[6] = ConfigData.StopBits;
           arrbyData[7] = (byte)(ConfigData.HardwareFlowControl ? 1 : 0);
           arrbyData[8] = (byte)(ConfigData.SendTestMessage ? 1 : 0);
           dwEomTemp = (UInt32)ConfigData.EOM1 << 24 + ConfigData.EOM2 << 16 + ConfigData.EOM3 << 8 + ConfigData.EOM4;

           // Write the first section
           WriteBlock((UInt32)(SIOMM.MM.SERIAL_MODULE.CONFIG_BASE + sizeof(UInt32) +
                                (SIOMM.MM.SERIAL_MODULE.CONFIG_BOUNDARY * SerialPort)),
                                arrbyData);

           // Write the second section (End-of-message characters)
           WriteQuad((UInt32)(SIOMM.MM.SERIAL_MODULE.EOM_BASE + (SIOMM.MM.SERIAL_MODULE.EOM_BOUNDARY * SerialPort)),
                            dwEomTemp);
       }


       public UInt64 GetScratchPadBitArea()
       //-------------------------------------------------------------------------------------------------
       // 
       //-------------------------------------------------------------------------------------------------
       {
           return GetBitmask64(SIOMM.MM.SCRATCHPAD.BITS_BASE);
       }


       public void SetScratchPadBitArea(UInt64 Bits)
       //-------------------------------------------------------------------------------------------------
       // 
       //-------------------------------------------------------------------------------------------------
       {
           SetBitmask64(SIOMM.MM.SCRATCHPAD.BITS_BASE, Bits);
       }


       public void SetScratchPadBitAreaMask(UInt64 OnMask, UInt64 OffMask)
       //-------------------------------------------------------------------------------------------------
       // 
       //-------------------------------------------------------------------------------------------------
       {
           byte[] arrbyData = new byte[16]; // buffer for the data to be read

           // Unpack the bytes
           O22SnapIoHelper.GetBytes(OnMask).CopyTo(arrbyData, 0);
           O22SnapIoHelper.GetBytes(OffMask).CopyTo(arrbyData, 8);

           // Write the data
           WriteBlock(SIOMM.MM.SCRATCHPAD.BITS_ON_MASK_BASE, arrbyData);
       }

       public UInt32 GetScratchPadInteger(int Index)
       {
           return GetScratchPadIntegerArea(Index, 1)[0];
       }


       public UInt32[] GetScratchPadIntegerArea(int StartIndex, int Length)
       //-------------------------------------------------------------------------------------------------
       // 
       //-------------------------------------------------------------------------------------------------
       {
           // Check the length
           if (Length > SIOMM.MM.SCRATCHPAD.INTEGER_MAX_ELEMENTS - StartIndex)
               Length = (int)SIOMM.MM.SCRATCHPAD.INTEGER_MAX_ELEMENTS - StartIndex;

           UInt32[] Data = new UInt32[Length];

           // Read the data
           byte[] arrbyData = ReadBlock((UInt32)(SIOMM.MM.SCRATCHPAD.INTEGER_BASE + (StartIndex * sizeof(UInt32))),
                               (UInt16)(Length * sizeof(UInt32)));

           // Unpack the data packet
           for (int i = 0; i < Length; i++)
               Data[i] = O22SnapIoHelper.ToUInt32(arrbyData, i * sizeof(UInt32));

           return Data;
       }

       public UInt32[] GetScratchPadIntegerAreaX(int StartIndex, int Length)
       //-------------------------------------------------------------------------------------------------
       // 
       //-------------------------------------------------------------------------------------------------
       {
           // Check the length
           if (Length > SIOMM.MM.SCRATCHPAD.INTEGER_MAX_ELEMENTS_X - StartIndex)
               Length = (int)SIOMM.MM.SCRATCHPAD.INTEGER_MAX_ELEMENTS_X - StartIndex;

           UInt32[] Data = new UInt32[Length];

           // Read the data
           byte[] arrbyData = ReadBlock((UInt32)(SIOMM.MM.SCRATCHPAD.INTEGER_BASE_X + (StartIndex * sizeof(UInt32))),
                               (UInt16)(Length * sizeof(UInt32)));

           // Unpack the data packet
           for (int i = 0; i < Length; i++)
               Data[i] = O22SnapIoHelper.ToUInt32(arrbyData, i * sizeof(UInt32));

           return Data;
       }

       public void SetScratchPadInteger(int Index, UInt32 Data)
       {
           SetScratchPadIntegerArea(Index, new uint[] { Data });
       }

       public void SetScratchPadIntegerX(int Index, UInt32 Data)
       {
           SetScratchPadIntegerAreaX(Index, new uint[] { Data });
       }

       public void SetScratchPadIntegerArea(int StartIndex, UInt32[] Data)
       //-------------------------------------------------------------------------------------------------
       // 
       //-------------------------------------------------------------------------------------------------
       {
           int Length = Data.Length;
           // Check the length
           if (Length > SIOMM.MM.SCRATCHPAD.INTEGER_MAX_ELEMENTS - StartIndex)
               Length = (int)SIOMM.MM.SCRATCHPAD.INTEGER_MAX_ELEMENTS - StartIndex;

           byte[] arrbyData = new byte[Length * sizeof(UInt32)]; // buffer for the data to be read

           // Pack the data packet
           for (int i = 0; i < Length; i++)
               O22SnapIoHelper.GetBytes(Data[i]).CopyTo(arrbyData, i * sizeof(UInt32));

           // Write the data
           WriteBlock((UInt32)(SIOMM.MM.SCRATCHPAD.INTEGER_BASE + (StartIndex * sizeof(UInt32))), arrbyData);
       }

       public void SetScratchPadIntegerAreaX(int StartIndex, UInt32[] Data)
       //-------------------------------------------------------------------------------------------------
       // 
       //-------------------------------------------------------------------------------------------------
       {
           int Length = Data.Length;
           // Check the length
           if (Length > SIOMM.MM.SCRATCHPAD.INTEGER_MAX_ELEMENTS_X - StartIndex)
               Length = (int)SIOMM.MM.SCRATCHPAD.INTEGER_MAX_ELEMENTS_X - StartIndex;

           byte[] arrbyData = new byte[Length * sizeof(UInt32)]; // buffer for the data to be read

           // Pack the data packet
           for (int i = 0; i < Length; i++)
               O22SnapIoHelper.GetBytes(Data[i]).CopyTo(arrbyData, i * sizeof(UInt32));

           // Write the data
           WriteBlock((UInt32)(SIOMM.MM.SCRATCHPAD.INTEGER_BASE_X + (StartIndex * sizeof(UInt32))), arrbyData);
       }

       public UInt64[] GetScratchPadInteger64Area(int StartIndex, int Length)
       //-------------------------------------------------------------------------------------------------
       // 
       //-------------------------------------------------------------------------------------------------
       {
           // Check the length
           if (Length > SIOMM.MM.SCRATCHPAD.INTEGER64_MAX_ELEMENTS - StartIndex)
               Length = (int)SIOMM.MM.SCRATCHPAD.INTEGER64_MAX_ELEMENTS - StartIndex;

           UInt64[] Data = new UInt64[Length];

           // Read the data
           byte[] arrbyData = ReadBlock((UInt32)(SIOMM.MM.SCRATCHPAD.INTEGER64_BASE + (StartIndex * sizeof(UInt64))),
                               (UInt16)(Length * sizeof(UInt64)));

           // Unpack the data packet
           for (int i = 0; i < Length; i++)
               Data[i] = O22SnapIoHelper.ToUInt64(arrbyData, i * sizeof(UInt64));

           return Data;
       }

       public void SetScratchPadInteger64Area(int StartIndex, UInt64[] Data)
       //-------------------------------------------------------------------------------------------------
       // 
       //-------------------------------------------------------------------------------------------------
       {
           int Length = Data.Length;
           // Check the length
           if (Length > SIOMM.MM.SCRATCHPAD.INTEGER64_MAX_ELEMENTS - StartIndex)
               Length = (int)SIOMM.MM.SCRATCHPAD.INTEGER64_MAX_ELEMENTS - StartIndex;

           byte[] arrbyData = new byte[Length * sizeof(UInt64)]; // buffer for the data to be read

           // Pack the data packet
           for (int i = 0; i < Length; i++)
               O22SnapIoHelper.GetBytes(Data[i]).CopyTo(arrbyData, i * sizeof(UInt64));

           // Write the data
           WriteBlock((UInt32)(SIOMM.MM.SCRATCHPAD.INTEGER64_BASE + (StartIndex * sizeof(UInt64))), arrbyData);
       }

       public Single[] GetScratchPadFloatArea(int StartIndex, int Length)
       //-------------------------------------------------------------------------------------------------
       // 
       //-------------------------------------------------------------------------------------------------
       {
           // Check the length
           if (Length > SIOMM.MM.SCRATCHPAD.FLOAT_MAX_ELEMENTS - StartIndex) 
               Length = (int)SIOMM.MM.SCRATCHPAD.FLOAT_MAX_ELEMENTS - StartIndex;

           Single[] Data = new Single[Length];

           // Read the data
           byte[] arrbyData = ReadBlock((UInt32)(SIOMM.MM.SCRATCHPAD.FLOAT_BASE + (StartIndex * sizeof(Single))),
                               (UInt16)(Length * sizeof(Single)));

           // Unpack the data packet
           for (int i = 0; i < Length; i++)
               Data[i] = O22SnapIoHelper.ToSingle(arrbyData, i * sizeof(Single));

           return Data;
       }

       public Single[] GetScratchPadFloatAreaX(int StartIndex, int Length)
       //-------------------------------------------------------------------------------------------------
       // 
       //-------------------------------------------------------------------------------------------------
       {
           // Check the length
           if (Length > SIOMM.MM.SCRATCHPAD.FLOAT_MAX_ELEMENTS_X - StartIndex)  
               Length = (int)SIOMM.MM.SCRATCHPAD.FLOAT_MAX_ELEMENTS_X - StartIndex;

           Single[] Data = new Single[Length];

           // Read the data
           byte[] arrbyData = ReadBlock((UInt32)(SIOMM.MM.SCRATCHPAD.FLOAT_BASE_X + (StartIndex * sizeof(Single))),
                               (UInt16)(Length * sizeof(Single)));

           // Unpack the data packet
           for (int i = 0; i < Length; i++)
               Data[i] = O22SnapIoHelper.ToSingle(arrbyData, i * sizeof(Single));

           return Data;
       }

       public void SetScratchPadFloatArea(int StartIndex, Single[] Data)
       //-------------------------------------------------------------------------------------------------
       // 
       //-------------------------------------------------------------------------------------------------
       {
           // Check the length
           int Length = Data.Length;
           if (Length > SIOMM.MM.SCRATCHPAD.FLOAT_MAX_ELEMENTS - StartIndex)
               Length = (int)SIOMM.MM.SCRATCHPAD.FLOAT_MAX_ELEMENTS - StartIndex;

           byte[] arrbyData = new byte[Length * sizeof(Single)]; // buffer for the data to be written

           // Pack the data packet
           for (int i = 0; i < Length; i++)
               O22SnapIoHelper.GetBytes(Data[i]).CopyTo(arrbyData, i * sizeof(Single));

           // Read the data
           WriteBlock((UInt32)(SIOMM.MM.SCRATCHPAD.FLOAT_BASE + (StartIndex * sizeof(Single))), arrbyData);
       }

       public void SetScratchPadFloatAreaX(int StartIndex, Single[] Data)
       //-------------------------------------------------------------------------------------------------
       // 
       //-------------------------------------------------------------------------------------------------
       {
           // Check the length
           int Length = Data.Length;
           if (Length > SIOMM.MM.SCRATCHPAD.FLOAT_MAX_ELEMENTS_X - StartIndex)
               Length = (int)SIOMM.MM.SCRATCHPAD.FLOAT_MAX_ELEMENTS_X - StartIndex;

           byte[] arrbyData = new byte[Length * sizeof(Single)]; // buffer for the data to be written

           // Pack the data packet
           for (int i = 0; i < Length; i++)
               O22SnapIoHelper.GetBytes(Data[i]).CopyTo(arrbyData, i * sizeof(Single));

           // Read the data
           WriteBlock((UInt32)(SIOMM.MM.SCRATCHPAD.FLOAT_BASE_X + (StartIndex * sizeof(Single))), arrbyData);
       }

       public String GetScratchPadString(int Index)
       {
           return GetScratchPadStringArea(Index, 1)[0];
       }

       public String[] GetScratchPadStringArea(int StartIndex, int Length)
       //-------------------------------------------------------------------------------------------------
       // 
       //-------------------------------------------------------------------------------------------------
       {
           // Check the length
           if (Length > SIOMM.MM.SCRATCHPAD.STRING_MAX_ELEMENTS - StartIndex) // 8 - now 64
               Length = (int)SIOMM.MM.SCRATCHPAD.STRING_MAX_ELEMENTS - StartIndex;

           String[] StringData = new String[Length];

           // Read the data
           byte[] arrbyData = ReadBlock((UInt32)(SIOMM.MM.SCRATCHPAD.STRING_BASE + (StartIndex * SIOMM.MM.SCRATCHPAD.STRING_BOUNDARY)),
                               (UInt16)(Length * SIOMM.MM.SCRATCHPAD.STRING_BOUNDARY));

           // Unpack the data packet
           for (int i = 0; i < Length; i++)
           {
               UInt16 wLength = O22SnapIoHelper.ToUInt16(arrbyData, (int)(i * SIOMM.MM.SCRATCHPAD.STRING_BOUNDARY));
               
               System.Text.ASCIIEncoding ae = new ASCIIEncoding();
               StringData[i] = ae.GetString(arrbyData,
                   (int)((i * SIOMM.MM.SCRATCHPAD.STRING_BOUNDARY) + SIOMM.MM.SCRATCHPAD.STRING_LENGTH_SIZE),
                   wLength); //.Substring(0, (int)SIOMM.MM.SCRATCHPAD.STRING_DATA_SIZE);
           }

           return StringData;
       }

       public void SetScratchPadString(int Index, String Data)
       {
           SetScratchPadStringArea(Index, new String[] {Data} );
       }

       public void SetScratchPadStringArea(int StartIndex, String[] StringData)
       //-------------------------------------------------------------------------------------------------
       // 
       //-------------------------------------------------------------------------------------------------
       {
           // Check the length
           int Length = StringData.Length;
           if (Length > SIOMM.MM.SCRATCHPAD.STRING_MAX_ELEMENTS - StartIndex)
               Length = (int)SIOMM.MM.SCRATCHPAD.STRING_MAX_ELEMENTS - StartIndex;

           byte[] arrbyData = new byte[SIOMM.MM.SCRATCHPAD.STRING_BOUNDARY * Length]; // buffer 

           // Pack the data packet
           for (int i = 0; i < Length; i++)
           {
               UInt16 TempLength = (UInt16)StringData[i].Length;
               if (TempLength > SIOMM.MM.SCRATCHPAD.STRING_DATA_SIZE)
               {
                   StringData[i] = StringData[i].Substring(0, (int)SIOMM.MM.SCRATCHPAD.STRING_DATA_SIZE);
                   TempLength = (UInt16)SIOMM.MM.SCRATCHPAD.STRING_DATA_SIZE;
               }

               O22SnapIoHelper.GetBytes(TempLength).CopyTo(arrbyData, i * SIOMM.MM.SCRATCHPAD.STRING_BOUNDARY);
               //byte[] rv = new byte[StringData[i].Length];

               System.Text.ASCIIEncoding ae = new ASCIIEncoding();
               ae.GetBytes(StringData[i]).CopyTo(arrbyData, (i * SIOMM.MM.SCRATCHPAD.STRING_BOUNDARY) + SIOMM.MM.SCRATCHPAD.STRING_LENGTH_SIZE);
           }

           // Write the data
           WriteBlock((UInt32)(SIOMM.MM.SCRATCHPAD.STRING_BASE + (StartIndex * SIOMM.MM.SCRATCHPAD.STRING_BOUNDARY)), arrbyData);
       }
        
    }

}
